#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Define.

??? Example "Define Examples"
    Usage:

        >>> from tabulate import tabulate
        >>> import ucdp as u
        >>> u.Define("param_p")
        Define('param_p')

        Complex types are NOT supported.

        >>> param = u.Define("param_p")
        >>> for item in param:
        ...     print(repr(item))
        Define('param_p')
"""

from typing import Any, ClassVar

from .consts import PAT_IDENTIFIER
from .doc import Doc
from .namespace import Namespace
from .nameutil import split_suffix
from .object import Field, Light, NamedObject, PosArgs


class Define(NamedObject, Light):
    """Define.

    Args:
        name: Name.

    Attributes:
        doc: Documentation Container
        value (Any): Value.

    ??? Example "Define Examples"
        Example:

            >>> import ucdp as u
            >>> cnt = u.Define("cnt_p")
            >>> cnt
            Define('cnt_p')
            >>> cnt.name
            'cnt_p'
            >>> cnt.basename
            'cnt'
            >>> cnt.suffix
            '_p'
            >>> cnt.doc
            Doc()
            >>> cnt.value

        If the parameter is casted via `int()` it returns `value` if set, other `type_.default`.

            >>> int(u.Define("cnt_p"))
            0
            >>> int(u.Define("cnt_p", value=4))
            4

        Define are Singleton:

            >>> u.Define("cnt_p") is u.Define("cnt_p")
            True
    """

    name: str = Field(pattern=PAT_IDENTIFIER)
    doc: Doc = Doc()
    value: Any = None

    _posargs: ClassVar[PosArgs] = ("name",)

    def __init__(self, name: str, **kwargs):
        super().__init__(name=name, **kwargs)  # type: ignore[call-arg]

    @property
    def basename(self):
        """Base Name."""
        return split_suffix(self.name)[0]

    @property
    def suffix(self):
        """Suffix."""
        return split_suffix(self.name)[1]

    def __str__(self) -> str:
        return self.name

    def __int__(self):
        return int(self.value or 0)

    def __iter__(self):
        yield self


class Defines(Namespace):
    """Defines."""



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Module Iterator.

Module Iteration Strategies:

* :any:`ModPreIter` - yield top module **before** the child-modules.
* :any:`ModPostIter` - yield top module **after** the child-modules.
"""

from abc import abstractmethod
from collections.abc import Callable, Iterable, Iterator
from itertools import chain
from typing import TypeAlias

import uniquer

from .iterutil import Names, namefilter
from .modbase import BaseMod
from .object import Object

FilterFunc: TypeAlias = Callable[[BaseMod], bool]
StopFunc: TypeAlias = Callable[[BaseMod], bool]
MaxLevel: TypeAlias = int


def _no_filter(mod: BaseMod):
    return True


def _no_stop(mod: BaseMod):
    return False


class BaseModIter(Object):
    """Base Class for Module Iterators."""

    mod: BaseMod
    filter_: FilterFunc | None = None
    stop: StopFunc | None = None
    stop_insts: StopFunc | None = None
    maxlevel: MaxLevel | None = None
    unique: bool = False

    def __init__(self, mod: BaseMod, **kwargs):
        super().__init__(mod=mod, **kwargs)  # type: ignore[call-arg]

    def __iter__(self):
        mod = self.mod
        filter_ = self.filter_ or _no_filter
        stop = self.stop or _no_stop
        stop_insts = self.stop_insts or _no_stop
        maxlevel = self.maxlevel or -1
        iter_ = self._iter([mod], filter_, stop, stop_insts, maxlevel - 1)
        if self.unique:
            iter_ = uniquemods(iter_)
        return iter_

    @staticmethod
    @abstractmethod
    def _iter(
        mods: Iterable[BaseMod], filter_: FilterFunc, stop: StopFunc, stop_insts: StopFunc, maxlevel: MaxLevel
    ) -> Iterator[BaseMod]:
        pass  # pragma: no cover


class ModPreIter(BaseModIter):
    """

    Iterate over module hierarchy starting at `mod`, using the pre-order strategy.

    Yield top module **before** the child-modules.

    Attributes:
        filter_: function called with every `mod` as argument, `mod` is returned if `True`.
        stop: stop iteration at `mod` if `stop` function returns `True` for `mod`.
        maxlevel (int): maximum descending in the mod hierarchy.
        unique (bool): Just return module once.

    ??? Example "Module Pre Iterator Examples"
        pre-order strategy Examples.

            >>> import ucdp as u
            >>> class CMod(u.AMod):
            ...     def _build(self) -> None:
            ...         pass
            >>> class EMod(u.AMod):
            ...     def _build(self) -> None:
            ...         pass
            >>> class AMod(u.AMod):
            ...     def _build(self) -> None:
            ...         pass
            >>> class DMod(u.AMod):
            ...     def _build(self) -> None:
            ...         CMod(self, "c")
            ...         EMod(self, "e")
            >>> class BMod(u.AMod):
            ...     def _build(self) -> None:
            ...         AMod(self, "a")
            ...         DMod(self, "d")
            >>> class HMod(u.AMod):
            ...     def _build(self) -> None:
            ...         pass
            >>> class IMod(u.AMod):
            ...     def _build(self) -> None:
            ...         HMod(self, "h")
            >>> class GMod(u.AMod):
            ...     def _build(self) -> None:
            ...         IMod(self, "i")
            >>> class FMod(u.AMod):
            ...     def _build(self) -> None:
            ...         BMod(self, "b")
            ...         GMod(self, "g")

            >>> f = FMod(None, "f")
            >>> [mod.name for mod in ModPreIter(f)]
            ['f', 'b', 'a', 'd', 'c', 'e', 'g', 'i', 'h']
            >>> [mod.name for mod in ModPreIter(f, maxlevel=3)]
            ['f', 'b', 'a', 'd', 'g', 'i']
            >>> [mod.name for mod in ModPreIter(f, maxlevel=3, filter_=lambda n: n.name not in 'eg')]
            ['f', 'b', 'a', 'd', 'i']
            >>> [mod.name for mod in ModPreIter(f, maxlevel=3, filter_=lambda n: n.name not in 'eg',
            ...                                 stop=lambda n: n.name == 'd')]
            ['f', 'b', 'a', 'i']
            >>> [mod.name for mod in ModPreIter(f, maxlevel=3, stop=lambda n: n.name == 'd')]
            ['f', 'b', 'a', 'g', 'i']
            >>> [mod.name for mod in ModPreIter(f, filter_=lambda n: n.name not in 'eg')]
            ['f', 'b', 'a', 'd', 'c', 'i', 'h']
            >>> [mod.name for mod in ModPreIter(f, stop=lambda n: n.name == 'd')]
            ['f', 'b', 'a', 'g', 'i', 'h']
            >>> [mod.name for mod in ModPreIter(f, filter_=lambda n: n.name not in 'eg', stop=lambda n: n.name == 'd')]
            ['f', 'b', 'a', 'i', 'h']
            >>> [mod.name for mod in ModPreIter(f)]
            ['f', 'b', 'a', 'd', 'c', 'e', 'g', 'i', 'h']
            >>> [mod.modname for mod in ModPreIter(f, unique=True)]
            ['f', 'b', 'a', 'd', 'c', 'e', 'g', 'i', 'h']
            >>> [mod.name for mod in ModPreIter(f, stop_insts=lambda n: n.name == 'b')]
            ['f', 'b', 'g', 'i', 'h']
    """

    @staticmethod
    def _iter(
        mods: Iterable[BaseMod], filter_: FilterFunc, stop: StopFunc, stop_insts: StopFunc, maxlevel: MaxLevel
    ) -> Iterator[BaseMod]:
        # TODO: heap impl
        for mod in mods:
            if stop(mod):
                continue
            if filter_(mod):
                yield mod
            if not stop_insts(mod) and maxlevel:
                yield from ModPreIter._iter(mod.insts, filter_, stop, stop_insts, maxlevel - 1)


class ModPostIter(BaseModIter):
    """
    Iterate over module hierarchy starting at `mod`, using the post-order strategy.

    Yield top module **after** the child-modules.

    Attributes:
        filter_ (FilterFunc): function called with every `mod` as argument, `mod` is returned if `True`.
        stop (StopFunc): stop iteration at `mod` if `stop` function returns `True` for `mod`.
        maxlevel (int): maximum descending in the mod hierarchy.
        unique (bool): Just return module once.

    ??? Example "Module Post Iterator Examples"
        pre-order strategy Examples.

            >>> import ucdp as u
            >>> class CMod(u.AMod):
            ...     def _build(self) -> None:
            ...         pass
            >>> class EMod(u.AMod):
            ...     def _build(self) -> None:
            ...         pass
            >>> class AMod(u.AMod):
            ...     def _build(self) -> None:
            ...         pass
            >>> class DMod(u.AMod):
            ...     def _build(self) -> None:
            ...         CMod(self, "c")
            ...         EMod(self, "e")
            >>> class BMod(u.AMod):
            ...     def _build(self) -> None:
            ...         AMod(self, "a")
            ...         DMod(self, "d")
            >>> class HMod(u.AMod):
            ...     def _build(self) -> None:
            ...         pass
            >>> class IMod(u.AMod):
            ...     def _build(self) -> None:
            ...         HMod(self, "h")
            >>> class GMod(u.AMod):
            ...     def _build(self) -> None:
            ...         IMod(self, "i")
            >>> class FMod(u.AMod):
            ...     def _build(self) -> None:
            ...         BMod(self, "b")
            ...         GMod(self, "g")

            >>> f = FMod(None, "f")
            >>> [mod.name for mod in ModPostIter(f)]
            ['a', 'c', 'e', 'd', 'b', 'h', 'i', 'g', 'f']
            >>> [mod.name for mod in ModPostIter(f, maxlevel=3)]
            ['a', 'd', 'b', 'i', 'g', 'f']
            >>> [mod.name for mod in ModPostIter(f, maxlevel=3, filter_=lambda n: n.name not in 'eg')]
            ['a', 'd', 'b', 'i', 'f']
            >>> [mod.name for mod in ModPostIter(f, maxlevel=3, filter_=lambda n: n.name not in 'eg',
            ...                                  stop=lambda n: n.name == 'd')]
            ['a', 'b', 'i', 'f']
            >>> [mod.name for mod in ModPostIter(f, maxlevel=3, stop=lambda n: n.name == 'd')]
            ['a', 'b', 'i', 'g', 'f']
            >>> [mod.name for mod in ModPostIter(f, filter_=lambda n: n.name not in 'eg')]
            ['a', 'c', 'd', 'b', 'h', 'i', 'f']
            >>> [mod.name for mod in ModPostIter(f, stop=lambda n: n.name == 'd')]
            ['a', 'b', 'h', 'i', 'g', 'f']
            >>> [mod.name for mod in ModPostIter(f, filter_=lambda n: n.name not in 'eg', stop=lambda n: n.name == 'd')]
            ['a', 'b', 'h', 'i', 'f']
            >>> [mod.name for mod in ModPostIter(f)]
            ['a', 'c', 'e', 'd', 'b', 'h', 'i', 'g', 'f']
            >>> [mod.modname for mod in ModPostIter(f, unique=True)]
            ['a', 'c', 'e', 'd', 'b', 'h', 'i', 'g', 'f']
            >>> [mod.name for mod in ModPostIter(f, stop_insts=lambda n: n.name == 'b')]
            ['b', 'h', 'i', 'g', 'f']
    """

    @staticmethod
    def _iter(
        mods: Iterable[BaseMod], filter_: FilterFunc, stop: StopFunc, stop_insts: StopFunc, maxlevel: MaxLevel
    ) -> Iterator[BaseMod]:
        # TODO: heap impl
        for mod in mods:
            if stop(mod):
                continue
            if not stop_insts(mod) and maxlevel:
                yield from ModPostIter._iter(mod.insts, filter_, stop, stop_insts, maxlevel - 1)
            if filter_(mod):
                yield mod


def get_mod(topmod: BaseMod, namepats: Names, base=False) -> BaseMod:
    """
    Return the one and just the one hardware module matching `namepats`.

    Iterate over `topmod` and all its submodules and return matching one.

    Args:
        topmod: Top module instance

    Parameter:
        namepats: Iterable with name pattern (including `*` and `?`) or comma separated string.
        base: namepats must match against module `basequalnames` instead of `qual_name`.
    """
    mods = get_mods(topmod, namepats, unique=True, base=base)
    if len(mods) == 1:
        return mods[0]
    listed_mods = mods or ModPostIter(topmod, unique=True)
    if base:
        names = sorted(uniquer.unique(chain.from_iterable(mod.basequalnames for mod in listed_mods)))
    else:
        names = sorted(uniquer.unique(mod.qualname for mod in listed_mods))
    if mods:
        lines = (f"Found multiple hardware modules for {namepats!r}:", *names)
    else:
        lines = (f"{namepats!r} not found. Known are:", *names)
    raise ValueError("\n  ".join(lines))


def get_mods(
    topmod: BaseMod, namepats: Names | None = None, unique: bool = False, base: bool = False
) -> tuple[BaseMod, ...]:
    """
    Return all modules matching `namepats` from hierarchy of `topmod`.

    Iterate over `topmod` and all its submodules and return matching ones.

    Args:
        topmod: Top module instance

    Keyword Args:
        namepats: Iterable with name pattern (including `*` and `?`) or comma separated string. All by default.
        unique (bool): Just return every module once.
        base: namepats must match against module `basequalnames` instead of `qual_name`.
    """
    if namepats:
        patfilter = namefilter(namepats)

        if base:

            def filter_(mod):
                qualnames = uniquer.unique(mod.basequalnames)
                return any(patfilter(qualname) for qualname in qualnames)
        else:

            def filter_(mod):
                return patfilter(mod.qualname)

        return tuple(ModPostIter(topmod, filter_=filter_, unique=unique))
    return tuple(ModPostIter(topmod, unique=unique))


def uniquemods(mods: Iterable[BaseMod]) -> Iterator[BaseMod]:
    """Iterate over unique modules."""
    qualnames = set()
    for mod in mods:
        qualname = mod.qualname
        if qualname not in qualnames:
            qualnames.add(qualname)
            yield mod



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


"""
Driver Tracking.

"""

from functools import cached_property
from typing import Any

from pydantic import computed_field

from .drivers import Source
from .expr import ConcatExpr, ConstExpr, Expr
from .slices import Slice
from .typehelper import BaseScalarTypes

SliceMap = dict[Slice, Expr | None]


class SliceAssign(Expr):
    """Slice Assignment."""

    type_: BaseScalarTypes
    expr: Expr

    @computed_field
    @cached_property
    def _slicemap(self) -> SliceMap:
        return {
            Slice(width=self.type_.width): None,
        }

    def get_expr(self, default: Source | None) -> ConcatExpr:
        """Convert To Concatenate."""
        slicemap = self._slicemap
        exprs = []
        # Create default number in case of missing default or if default is a Note
        if default is None or not isinstance(default, Expr):
            default = ConstExpr(self.type_)
        for slice_, item in sorted(slicemap.items(), key=lambda pair: pair[0].right, reverse=True):
            if item is None:
                item = default[slice_]  # noqa: PLW2901
            exprs.append(item)
        return ConcatExpr(tuple(exprs))

    def set(self, slice_: Slice, expr: Expr, overwrite: bool):
        slicemap = self._slicemap
        width = self.type_.width
        if slice_.left >= width:
            raise ValueError(f"Slice {slice_} exceeds width ({width})")
        # Remove empty/overwritten slices
        for sslice in self._searchpadding(slicemap, slice_, overwrite):
            slicemap.pop(sslice)
        self._addpadding(slicemap, slice_.right - 1, sslice.right)
        slicemap[slice_] = expr
        self._addpadding(slicemap, sslice.left, slice_.left + 1)

    @staticmethod
    def _searchpadding(slicemap: SliceMap, slice_: Slice, overwrite: bool) -> list[Slice]:
        slices = []
        for islice, iitem in slicemap.items():
            if islice.mask & slice_.mask:
                if iitem and not overwrite:
                    raise ValueError(f"Slice {slice_} is already taken by {iitem}")
                slices.append(islice)
        return slices

    @staticmethod
    def _addpadding(slicemap: SliceMap, left: Any, right: Any):
        if left >= right:
            slice_ = Slice(left, right)
            slicemap[slice_] = None



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Code Generator based of FileLists."""

import importlib
import re
import sys
from collections.abc import Iterable
from concurrent.futures import ThreadPoolExecutor
from contextlib import AbstractContextManager, contextmanager
from pathlib import Path
from typing import Any

from makolator import Config, Datamodel, Existing, Makolator
from uniquer import uniquelist

from .cache import CACHE
from .consts import TEMPLATE_PATHS
from .filelistparser import FileListParser
from .logging import LOGGER
from .modbase import BaseMod
from .modfilelist import iter_modfilelists
from .object import Object
from .top import Top
from .util import extend_sys_path

Paths = Path | Iterable[Path]
Data = dict[str, Any]

_RE_PATH_EDITABLE = re.compile(r"__editable__\.([^-]+)-")


def get_template_paths(paths: Iterable[Path] | None = None) -> list[Path]:
    """
    Determine Template Paths.

    Keyword Args:
        paths: Search Path For Data Model And Template Files.
    """
    template_paths: list[Path] = []
    with extend_sys_path(paths, use_env_default=True):
        for path in sys.path:
            editable = _RE_PATH_EDITABLE.match(path)
            if editable:
                package_name = editable.group(1)
                spec = importlib.util.find_spec(package_name)
                if spec and spec.submodule_search_locations:
                    for location in spec.submodule_search_locations:
                        for pattern in TEMPLATE_PATHS:
                            add_paths = tuple(Path(location).glob(pattern))
                            LOGGER.debug(f"get_template_paths: editable={path} {location}/{pattern}: {add_paths}")
                            template_paths.extend(add_paths)
                else:
                    LOGGER.warning(f"Cannot determine template_path for editable path {path}")
            else:
                for pattern in TEMPLATE_PATHS:
                    add_paths = tuple(Path(path).glob(f"*/{pattern}"))
                    LOGGER.debug(f"get_template_paths: {path}/*/{pattern}: {add_paths}")
                    template_paths.extend(add_paths)
    return uniquelist(template_paths)


def get_makolator(
    show_diff: bool = False,
    verbose: bool = True,
    paths: Iterable[Path] | None = None,
    create: bool = False,
    force: bool | None = None,
) -> Makolator:
    """
    Create Makolator.

    Keyword Args:
        show_diff: Show Changes.
        verbose: Display updated files.
        paths: Search Path For Data Model And Template Files.
        create: Create missing inplace files.
        force: overwrite existing files.
    """
    diffout = print if show_diff else None
    template_paths = get_template_paths(paths=paths)
    if force is True:
        existing = Existing.OVERWRITE
    elif force is None:
        existing = Existing.KEEP_TIMESTAMP
    else:
        existing = Existing.KEEP

    config = Config(
        template_paths=template_paths,
        marker_linelength=80,
        diffout=diffout,
        verbose=verbose,
        cache_path=CACHE.templates_path,
        track=True,
        create=create,
        existing=existing,
    )
    return Makolator(config=config)


def get_datamodel(top: Top | BaseMod, data: Data | None = None) -> Datamodel:
    """Create Data Model."""
    datamodel = Datamodel()
    if isinstance(top, BaseMod):
        top = Top.from_mod(top)
    if data:
        datamodel.__dict__.update(data)
    datamodel.top = top
    return datamodel


class Generator(Object):
    """Generator."""

    makolator: Makolator
    maxworkers: int | None = None
    check: bool = False
    no_stat: bool = False

    def __enter__(self) -> "Generator":
        LOGGER.debug("%s", self.makolator.config)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        makolator = self.makolator
        config = makolator.config
        tracker = makolator.tracker
        if not self.no_stat and config.verbose and config.track:
            print(tracker.stat)
        if self.check:
            assert config.track
            if tracker.total != tracker.identical:
                modified = tracker.total - tracker.identical
                LOGGER.error("%d files modified", modified)

    @contextmanager
    def top(
        self,
        top: Top | BaseMod,
        data: Data | None = None,
    ) -> AbstractContextManager[Top]:
        """Initialize Data Model."""
        makolator = self.makolator
        makolator.datamodel = datamodel = get_datamodel(top, data=data)
        yield datamodel.top

    def generate(
        self,
        top: Top | BaseMod,
        name: str,
        target: str | None = None,
        paths: Iterable[Path] | None = None,
        filelistparser: FileListParser | None = None,
        maxlevel: int | None = None,
        data: Data | None = None,
        clean: bool = False,
    ):
        """
        Generate for Top-Module.

        Args:
            top: Top
            name: Filelist Name

        Keyword Args:
            target: Target Filter
            paths: Search Path For Data Model And Template Files.
            filelistparser: Specific File List Parser
            maxlevel: Stop Generation on given hierarchy level.
            data: Data added to the datamodel.
            clean: Remove obsolete fully-generated files.
        """
        with self.top(top, data=data) as top_:
            modfilelists = iter_modfilelists(
                top_.mod,
                name,
                target=target,
                filelistparser=filelistparser,
                replace_envvars=True,
                maxlevel=maxlevel,
            )
            with extend_sys_path(paths, use_env_default=True):
                self._generate(modfilelists, clean)

    def _generate(self, modfilelists, clean: bool) -> None:  # noqa: C901
        makolator = self.makolator

        def _inplace(template_filepaths, filepath, context):
            if not filepath.exists() and not makolator.config.create:
                LOGGER.error("Inplace file %r missing", str(filepath))
            else:
                makolator.inplace(template_filepaths, filepath, context=context)

        def _gen(template_filepaths, filepath, context):
            makolator.gen(template_filepaths, filepath, context=context)

        def _clean(filepath):
            if makolator.is_fully_generated(filepath):
                makolator.remove(filepath)

        with ThreadPoolExecutor(max_workers=self.maxworkers) as exe:
            jobs = []  # type: ignore [var-annotated]
            all_filepaths: list[Path] = []
            all_clean_filepaths: list[Path] = []
            for mod, modfilelist in modfilelists:
                gen = modfilelist.get_gen(mod, modfilelist.flavor)
                if gen == "no":
                    continue
                if gen == "custom":
                    jobs.append(exe.submit(modfilelist.generate, mod))
                    continue
                filepaths: tuple[Path, ...] = modfilelist.filepaths or ()  # type: ignore[assignment]
                template_filepaths: tuple[Path, ...] = modfilelist.template_filepaths or ()  # type: ignore[assignment]
                inc_filepaths: tuple[Path, ...] = modfilelist.inc_filepaths or ()  # type: ignore[assignment]
                inc_template_filepaths: tuple[Path, ...] = modfilelist.inc_template_filepaths or ()  # type: ignore[assignment]
                for filepath in modfilelist.clean_filepaths:
                    if filepath not in all_clean_filepaths:
                        all_clean_filepaths.append(filepath)
                ctx = {"mod": mod, "modfilelist": modfilelist}
                if gen == "inplace":
                    jobs.extend(exe.submit(_inplace, inc_template_filepaths, path, ctx) for path in inc_filepaths)
                    jobs.extend(exe.submit(_inplace, template_filepaths, path, ctx) for path in filepaths)
                else:
                    jobs.extend(exe.submit(_gen, inc_template_filepaths, path, ctx) for path in inc_filepaths)
                    jobs.extend(exe.submit(_gen, template_filepaths, path, ctx) for path in filepaths)
                    for filepath in filepaths + inc_filepaths:
                        if filepath not in all_filepaths:
                            all_filepaths.append(filepath)
            for clean_filepath in all_clean_filepaths:
                if clean_filepath in all_filepaths:
                    continue
                jobs.append(exe.submit(_clean, clean_filepath))
            for job in jobs:
                job.result()


def generate(
    top: Top | BaseMod,
    name: str,
    target: str | None = None,
    filelistparser: FileListParser | None = None,
    makolator: Makolator | None = None,
    maxlevel: int | None = None,
    maxworkers: int | None = None,
    paths: Iterable[Path] | None = None,
    data: Data | None = None,
    create: bool = False,
    clean: bool = False,
):
    """
    Generate for Top-Module.

    Args:
        top: Top
        name: Filelist Name

    Keyword Args:
        target: Target Filter
        filelistparser: Specific File List Parser
        makolator: Specific Makolator
        maxlevel: Stop Generation on given hierarchy level.
        maxworkers: Maximal Parallelism.
        paths: Search Path For Data Model And Template Files.
        data: Data added to the datamodel.
        create: Create missing inplace files.
        clean: Remove obsolete fully-generated files.
    """
    makolator = makolator or get_makolator(paths=paths, create=create)
    with Generator(makolator=makolator, maxworkers=maxworkers) as generator:
        generator.generate(
            top=top,
            name=name,
            target=target,
            filelistparser=filelistparser,
            maxlevel=maxlevel,
            data=data,
            clean=clean,
        )


def clean(
    top: Top | BaseMod,
    name: str,
    target: str | None = None,
    filelistparser: FileListParser | None = None,
    makolator: Makolator | None = None,
    maxlevel: int | None = None,
    maxworkers: int | None = None,
    paths: Iterable[Path] | None = None,
    dry_run: bool = False,
    data: Data | None = None,
):
    """
    Remove Generated Files for Top-Module.

    Args:
        top: Top
        name: Filelist Name

    Keyword Args:
        target: Target Filter
        filelistparser: Specific File List Parser
        makolator: Specific Makolator
        maxlevel: Stop Generation on given hierarchy level.
        maxworkers: Maximal Parallelism.
        paths: Search Path For Data Model And Template Files.
        dry_run: Do nothing.
        data: Data added to the datamodel.
    """
    makolator = makolator or get_makolator(paths=paths)
    with Generator(makolator=makolator, no_stat=True) as generator:
        with generator.top(top, data=data) as top_:
            modfilelists = iter_modfilelists(
                top_.mod,
                name,
                target=target,
                filelistparser=filelistparser,
                replace_envvars=True,
                maxlevel=maxlevel,
            )
            with extend_sys_path(paths, use_env_default=True):
                with ThreadPoolExecutor(max_workers=maxworkers) as executor:
                    jobs = []
                    for _, modfilelist in modfilelists:
                        filepaths: tuple[Path, ...] = modfilelist.filepaths or ()  # type: ignore[assignment]
                        if modfilelist.gen == "full":
                            for filepath in filepaths:
                                print(f"Removing '{filepath!s}'")
                                if not dry_run:
                                    jobs.append(executor.submit(filepath.unlink, missing_ok=True))
                    for job in jobs:
                        job.result()
                if dry_run:
                    print("DRY RUN. Nothing done.")


def render_generate(
    top: Top | BaseMod,
    template_filepaths: Paths,
    genfile: Path | None = None,
    makolator: Makolator | None = None,
    paths: Iterable[Path] | None = None,
    data: Data | None = None,
    no_stat: bool = False,
):
    """
    Render Template and Generate File.

    Args:
        top: Top
        template_filepaths: Template Paths

    Keyword Args:
        genfile: Generated File
        makolator: Specific Makolator
        paths: Search Path For Data Model And Template Files.
        data: Data added to the datamodel.
        no_stat: Skip Statistics
    """
    makolator = makolator or get_makolator(paths=paths)
    with Generator(makolator=makolator, no_stat=no_stat) as generator:
        with generator.top(top, data=data):
            with extend_sys_path(paths, use_env_default=True):
                makolator.gen(template_filepaths, dest=genfile)


def render_inplace(
    top: Top | BaseMod,
    template_filepaths: Paths,
    inplacefile: Path,
    makolator: Makolator | None = None,
    ignore_unknown: bool = False,
    paths: Iterable[Path] | None = None,
    data: Data | None = None,
    no_stat: bool = False,
):
    """
    Render Template and Update File.

    Args:
        top: Top
        template_filepaths: Template Paths
        inplacefile: Updated File

    Keyword Args:
        makolator: Specific Makolator
        data: Data added to the datamodel.
        paths: Search Path For Data Model And Template Files.
        ignore_unknown: Ignore unknown inplace markers, instead of raising an error.
        no_stat: Skip Statistics
    """
    makolator = makolator or get_makolator(paths=paths)
    with Generator(makolator=makolator, no_stat=no_stat) as generator:
        with generator.top(top, data=data):
            with extend_sys_path(paths, use_env_default=True):
                makolator.inplace(template_filepaths, filepath=inplacefile, ignore_unknown=ignore_unknown)



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""IC Design Related Slice Handling."""

import enum
import re
from typing import Any

from .object import LightObject

_RE_STRING = re.compile(r"\[?(?P<left>[^:\]]+)(:(?P<right>[^\]]+))?\]?")


class SliceDirection(enum.Enum):
    """Slice Direction."""

    DOWN = 0
    UP = 1

    def __repr__(self):
        return self.name


DOWN = SliceDirection.DOWN
UP = SliceDirection.UP


class Slice(LightObject):
    """
    Bit slice of `width` bits starting at bit position `left` or `right`.

    >>> s = Slice(right=6, left=9)
    >>> s
    Slice('9:6')
    >>> s.left
    9
    >>> s.right
    6
    >>> s.width
    4
    >>> str(s)
    '9:6'
    >>> s.mask
    960
    >>> s.direction
    DOWN
    >>> s.slice
    slice(9, 6, -1)
    >>> tuple(s)
    (9, 8, 7, 6)
    >>> s.prev
    5
    >>> s.nxt
    10

    >>> s = Slice(left=6, right=9)
    >>> s
    Slice('6:9')
    >>> s.left
    6
    >>> s.right
    9
    >>> s.width
    4
    >>> str(s)
    '6:9'
    >>> s.mask
    960
    >>> s.direction
    UP
    >>> s.slice
    slice(6, 9, 1)
    >>> tuple(s)
    (6, 7, 8, 9)
    >>> s.prev
    5
    >>> s.nxt
    10

    >>> Slice(left=7, right=4) in Slice(left=7, right=4)
    True
    >>> Slice(left=7, right=5) in Slice(left=7, right=4)
    True
    >>> Slice(left=6, right=4) in Slice(left=7, right=4)
    True
    >>> Slice(left=7, right=4) in Slice(left=6, right=4)
    False
    >>> Slice(left=7, right=4) in Slice(left=7, right=5)
    False
    >>> Slice(left=7, right=5) in Slice(left=4, right=7)
    False

    >>> Slice('2:1')
    Slice('2:1')
    >>> Slice('1:2')
    Slice('1:2')
    >>> Slice(2)
    Slice('2')
    >>> Slice(right=2)
    Slice('2')
    >>> Slice(right=2, left=3)
    Slice('3:2')
    >>> Slice.cast(slice(2, 1))
    Slice('2:1')
    >>> Slice.cast(slice(1, 2))
    Slice('1:2')
    >>> Slice('')
    Traceback (most recent call last):
      ...
    ValueError: Invalid Slice Specification ''
    """

    left: Any
    right: Any
    width: Any

    def __init__(  # noqa: C901, PLR0912
        self,
        left: Any | None = None,
        right: Any | None = None,
        width: Any | None = None,
        direction: SliceDirection | None = None,
    ) -> None:
        if isinstance(left, str):
            if right is not None:
                raise ValueError("'right' must be None")
            if width is not None:
                raise ValueError("'width' must be None")
            mat = _RE_STRING.match(left)
            if mat:
                left = int(mat.group("left"))
                right = int(mat.group("right")) if mat.group("right") else None
            else:
                raise ValueError(f"Invalid Slice Specification {left!r}") from None

        # this is quite complex here - but left, right and width may be u.Expr and we want to have them minimal
        if width is not None:
            if left is None:
                # 'width' given
                if direction in (None, SliceDirection.DOWN):
                    # downwards
                    if right is None:
                        right = 0
                    if isinstance(width, int) and width == 1:
                        left = right
                    elif isinstance(right, int) and right == 0:
                        left = width - 1
                    else:
                        left = width - 1 + right
                elif right is None:
                    left = 0
                    right = width - 1
                else:
                    left = right - (width - 1)

            elif right is None:
                # left and width given
                if direction in (None, SliceDirection.DOWN):
                    # downwards
                    right = left - (width - 1)
                else:
                    # upwards
                    right = left + (width - 1)
            else:
                # 'left', 'right' and 'width' given
                raise ValueError("'left', 'right' AND 'width' given, this is one too much")

        elif left is None:
            left = right
            width = 1
        elif right is None:
            right = left
            width = 1
        else:
            width = self._calc_width(left, right)
        _check_direction(left, right, direction)
        super().__init__(left=left, right=right, width=width)  # type: ignore[call-arg]

    @staticmethod
    def _calc_width(left: Any, right: Any) -> Any:
        """Slice Width."""
        return abs(left - right) + 1

    @staticmethod
    def cast(value, direction: SliceDirection | None = None) -> "Slice":
        """
        Create :any:`Slice` from `value`.

        These three formats are supported:

        >>> Slice.cast("[15:4]")
        Slice('15:4')
        >>> Slice.cast("[4:15]")
        Slice('4:15')
        >>> Slice.cast("[16]")
        Slice('16')
        >>> Slice.cast(range(4,16))
        Slice('4:15')
        >>> Slice.cast(range(15, 3, -1))
        Slice('15:4')
        >>> Slice.cast('16')
        Slice('16')
        >>> Slice.cast(16)
        Slice('16')
        >>> Slice.cast(Slice('16'))
        Slice('16')
        >>> Slice.cast('')
        Traceback (most recent call last):
          ...
        ValueError: Invalid Slice Specification ''
        >>> Slice.cast(None)
        Traceback (most recent call last):
          ...
        ValueError: Invalid Slice Specification None
        >>> Slice.cast("[4]", direction=DOWN)
        Slice('4')
        >>> Slice.cast("[4:15]", direction=DOWN)
        Traceback (most recent call last):
          ...
        ValueError: Slice must be downwards but is upwards
        """
        slice_ = None
        if isinstance(value, Slice):
            slice_ = value
        elif isinstance(value, slice):
            slice_ = Slice(left=value.start, right=value.stop)
        elif isinstance(value, range):
            values = tuple(value)
            slice_ = Slice(left=values[0], right=values[-1])
        elif isinstance(value, int):
            slice_ = Slice(left=value)
        elif isinstance(value, str):
            mat = _RE_STRING.match(value)
            if mat:
                left = int(mat.group("left"))
                right = int(mat.group("right")) if mat.group("right") else None
                slice_ = Slice(left=left, right=right, direction=direction)
        if slice_ is not None:
            _check_direction(slice_.left, slice_.right, direction)
            return slice_
        raise ValueError(f"Invalid Slice Specification {value!r}") from None

    @property
    def bits(self):
        """
        Colon separated bits.

        >>> Slice(left=4, right=8).bits
        '4:8'
        >>> Slice(left=8, right=4).bits
        '8:4'
        >>> Slice(left=4).bits
        '4'
        >>> Slice(right=4).bits
        '4'
        """
        if self.width > 1:
            return f"{self.left}:{self.right}"
        return f"{self.left}"

    @property
    def prev(self):
        """Previous Bit respecting direction."""
        if self.left >= self.right:
            return self.right - 1
        return self.left - 1

    @property
    def nxt(self):
        """Next Free Bit respecting direction."""
        if self.left >= self.right:
            return self.right + self.width
        return self.left + self.width

    def __str__(self):
        return self.bits

    def __repr__(self):
        if self.width > 1:
            return f"{self.__class__.__qualname__}('{self.left}:{self.right}')"
        return f"{self.__class__.__qualname__}('{self.left}')"

    @property
    def mask(self):
        """
        Mask.

        >>> Slice(left=4, right=8).mask
        496
        >>> Slice(left=8, right=4).mask
        496
        >>> Slice(left=4).mask
        16
        >>> Slice(right=4).mask
        16
        """
        return ((2**self.width) - 1) << min(self.right, self.left)

    @property
    def direction(self) -> SliceDirection | None:
        """
        Direction.

        >>> Slice(left=4, right=8).direction
        UP
        >>> Slice(left=8, right=4).direction
        DOWN
        >>> Slice(left=4).direction
        >>> Slice(right=4).direction
        """
        return _get_direction(self.left, self.right)

    def extract(self, word: int, is_signed: bool = False) -> int:
        """
        Extract slice value from `word`.

        >>> slice = Slice(width=4, right=4)
        >>> hex(slice.mask)
        '0xf0'
        >>> slice.extract(0x193)
        9
        >>> slice.extract(0x193, is_signed=True)
        -7
        """
        value = (word & self.mask) >> self.right
        if is_signed:
            value = _unsigned_to_signed(value, self.width)
        return value

    def update(self, word, value, is_signed: bool = False):
        """
        Extract slice value from `word`.

        >>> slice = Slice(width=4, right=4)
        >>> hex(slice.mask)
        '0xf0'
        >>> hex(slice.update(0x123, 9))
        '0x193'
        >>> hex(slice.update(0x123, -7, is_signed=True))
        '0x193'
        >>> slice.update(0x123, -7)
        Traceback (most recent call last):
          ...
        ValueError: -7 is not a unsigned 4 bit integer
        """
        mask = self.mask
        if is_signed:
            value = _signed_to_unsigned(value, self.width)
        else:
            _check_unsigned(value, self.width)
        return (word & ~mask) | ((value << self.right) & mask)

    @property
    def slice(self):
        """Python Slice Equivalent."""
        step = -1 if self.left > self.right else 1
        return slice(self.left, self.right, step)

    def __contains__(self, other):
        if isinstance(other, Slice):
            direction = self.direction
            otherdirection = other.direction
            if direction and otherdirection and direction != otherdirection:
                return False
            mask = self.mask
            return mask == mask | other.mask
        return NotImplemented

    def __iter__(self):
        if self.left > self.right:
            yield from range(self.left, self.right - 1, -1)
        else:
            yield from range(self.left, self.right + 1, 1)


def _get_direction(left: Any, right: Any) -> SliceDirection | None:
    if left > right:
        return SliceDirection.DOWN
    if left < right:
        return SliceDirection.UP
    return None


def _check_direction(left: Any, right: Any, direction: SliceDirection | None):
    slicedirection = _get_direction(left, right)
    if direction and slicedirection and direction != slicedirection:
        req = direction.name.lower()
        act = slicedirection.name.lower()
        raise ValueError(f"Slice must be {req}wards but is {act}wards")


def _signed_to_unsigned(value: int, width: int) -> int:
    high = (1 << (width - 1)) - 1
    low = ~high
    if value < low or value > high:
        msg = f"{value} is not a signed {width} bit integer"
        raise ValueError(msg)
    return (value + (1 << width)) & (~(-1 << width))


def _check_unsigned(value: int, width: int) -> None:
    low = 0
    high = (1 << width) - 1
    if value < low or value > high:
        msg = f"{value} is not a unsigned {width} bit integer"
        raise ValueError(msg)


def _unsigned_to_signed(value: int, width: int) -> int:
    _check_unsigned(value, width)
    if value & (1 << (width - 1)):  # MSB set->negative
        return value - (1 << width)
    return value


def mask_to_slices(mask: int) -> tuple[Slice, ...]:
    """
    Convert `mask` to tuple of :any:`Slice` instances.

    >>> mask_to_slices(0x0FFFFFF0)
    (Slice('27:4'),)
    >>> mask_to_slices(0x0FFF0FF0)
    (Slice('11:4'), Slice('27:16'))
    """
    slices = []
    right = None
    for idx, bit in enumerate(reversed("0" + bin(mask)[2:])):
        if bit == "1":
            if right is None:
                right = idx
        elif right is not None:
            left = idx - 1
            slices.append(Slice(left=left, right=right))
            right = None
    return tuple(slices)



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Hardware Module.
"""

from .modbasetop import BaseTopMod
from .modutil import get_modname, get_topmodname


class AMod(BaseTopMod):
    """
    A Normal Module With Parameters And A Fixed Number Of Ports (maybe `ifdef` encapsulated).

    See [BaseMod][ucdp.modbase.BaseMod] for arguments, attributes and details.

    All module parameter, local parameter, ports, signals and submodules
    **MUST** be added and created within the `_build` method:


    ???+ Example "AMod Build Examples"
        _build(self):

            >>> import ucdp as u
            >>> class AdderMod(u.AMod):
            ...
            ...     def _build(self) -> None:
            ...         width_p = self.add_param(u.IntegerType(default=16), 'width_p')
            ...         datatype = u.UintType(width_p)
            ...         self.add_port(datatype, "a_i")
            ...         self.add_port(datatype, "b_i")
            ...         self.add_port(datatype, "y_o")

    Warning:
        There is no other build method on purpose!
    """

    @property
    def modname(self) -> str:
        """Module Name."""
        return get_modname(self.__class__)

    @property
    def topmodname(self) -> str:
        """Top Module Name."""
        return get_topmodname(self.__class__)



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Finder.
"""

from collections.abc import Iterator

from uniquer import unique

from ._modloader import (
    Paths,
    Patterns,
    TopModRefPat,
    build_top,
    find_modrefs,
    get_topmodrefpats,
    load_modcls,
)
from .iterutil import namefilter
from .logging import LOGGER
from .modbase import ModCls
from .modgenerictb import AGenericTbMod
from .moditer import get_mods
from .modref import ModRef
from .modtopref import TopModRef
from .modtoprefinfo import TopModRefInfo, is_top
from .util import extend_sys_path


def find(
    paths: Paths | None = None,
    patterns: Patterns | None = None,
    glob: bool = False,
    local: bool | None = None,
    is_top: bool | None = None,
) -> Iterator[TopModRefInfo]:
    """List All Available Module References."""
    pats = tuple(get_topmodrefpats(patterns))
    infos = unique(_find_infos(paths, pats, glob, local=local), key=lambda modrefinfo: str(modrefinfo.topmodref))
    if is_top is not None:
        infos = [info for info in infos if info.is_top == is_top]
    yield from sorted(infos, key=lambda modrefinfo: str(modrefinfo.topmodref))


def _find_infos(
    paths: Paths | None, pats: tuple[TopModRefPat | TopModRef, ...], glob: bool = False, local: bool | None = None
) -> Iterator[TopModRefInfo]:
    with extend_sys_path(paths, use_env_default=True):
        modrefs = find_modrefs(local=local)
        for pat in pats:
            mat = False
            if isinstance(pat, TopModRef):
                yield _create(pat)
                mat = True
            elif pat.tb:
                # testbench with top
                tbfilter = namefilter(pat.tb)
                for tbmodref in modrefs:
                    # filter by name
                    if not tbfilter(str(tbmodref)):
                        continue
                    # load tbmodcls
                    tbmodcls = load_modcls(tbmodref)
                    # skip non-generic testbenches
                    if not issubclass(tbmodcls, AGenericTbMod):
                        continue
                    # search tops
                    modclss = tuple(tbmodcls.dut_modclss)
                    for info in _find_tops(modrefs, pat.top, pat.sub, glob, modclss=modclss, tbmodref=tbmodref):
                        mat = True
                        yield info
            else:
                # top only
                for info in _find_tops(modrefs, pat.top, pat.sub, glob):
                    mat = True
                    yield info
            if not mat:
                LOGGER.error(f"{str(pat)!r} did not match any module")


def _find_tops(
    modrefs: tuple[ModRef, ...],
    toppat: str,
    subpat: str | None,
    glob: bool,
    modclss: tuple[ModCls, ...] | None = None,
    tbmodref: ModRef | None = None,
) -> Iterator[TopModRefInfo]:
    topfilter = namefilter(toppat)
    for topmodref in modrefs:
        # filter by name
        if not topfilter(str(topmodref)):
            continue
        topmodcls = load_modcls(topmodref)
        if subpat:
            # skip non-top
            if not is_top(topmodcls):
                continue
            # sub
            try:
                topmod = build_top(topmodcls)
            except Exception:  # noqa: S112
                continue
            for inst in get_mods(topmod, subpat):
                # skip modules, which should not be tested
                if modclss and not isinstance(inst, modclss):
                    continue
                yield _create(TopModRef(tb=tbmodref, top=topmodref, sub=inst.qualname))
        else:
            # skip modules, which should not be tested
            if not glob and modclss and not issubclass(topmodcls, modclss):
                continue
            yield _create(TopModRef(tb=tbmodref, top=topmodref))


def _create(topmodref: TopModRef) -> TopModRefInfo:
    tbmodcls = load_modcls(topmodref.tb) if topmodref.tb else None
    modcls = load_modcls(topmodref.top)
    return TopModRefInfo.create(topmodref, modcls, tbmodcls=tbmodcls)



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Structure Types.

A structure assembles multiple type, name, orientation pairs.

* :any:`StructItem` - Struct item
* :any:`StructType` - Standard Struct
* :any:`AAGlobalStructType` - A public struct which fills up through all instances.
* :any:`DynamicStructType` - A public struct which fills up per instance.
"""

from abc import abstractmethod
from collections.abc import Callable
from typing import Any, ClassVar

from .clkrelbase import BaseClkRel
from .consts import PAT_IDENTIFIER
from .dict import Dict
from .doc import Doc
from .docutil import doc_from_type
from .exceptions import LockError
from .ifdef import Ifdefs, cast_ifdefs
from .object import Field, Light, Object, PosArgs, PrivateField
from .orientation import BWD, FWD, Orientation
from .typebase import ACompositeType, BaseType


class StructItem(Object):
    """
    Struct NamedObject.

    Attributes:
        name: Name of struct member
        type_: Type of struct member
        orientation: Orientation of struct member. `FWD` by default
        doc: Documentation Container
        ifdefs: IFDEF encapsulations
        clkrel: Clock Relation
    """

    name: str = Field(pattern=PAT_IDENTIFIER)
    type_: BaseType
    orientation: Orientation = FWD
    doc: Doc = Doc()
    ifdefs: Ifdefs = ()
    clkrel: BaseClkRel | None = None

    _posargs: ClassVar[PosArgs] = ("name", "type_")

    def __init__(self, name, type_, **kwargs):
        super().__init__(name=name, type_=type_, **kwargs)

    @property
    def title(self):
        """Alias to `doc.title`."""
        return self.doc.title

    @property
    def descr(self):
        """Alias to `doc.descr`."""
        return self.doc.descr

    @property
    def comment(self):
        """Alias to `doc.comment`."""
        return self.doc.comment


StructFilter = Callable[[StructItem], bool]


class BaseStructType(Dict, ACompositeType):
    """Base Type for all Structs."""

    filter_: StructFilter | None = None
    _items: dict[Any, Any] = PrivateField(default_factory=dict)
    _is_locked: bool = PrivateField(default=False)

    def _add(
        self,
        name: str,
        type_: BaseType,
        orientation: Orientation = FWD,
        title: str | None = None,
        descr: str | None = None,
        comment: str | None = None,
        ifdef: str | None = None,
        ifdefs: Ifdefs | str | None = None,
        clkrel: BaseClkRel | None = None,
    ) -> None:
        """
        Add member `name` with `type_` and `orientation`.

        Args:
            name: Name
            type_: Type
            orientation: Orientation

        Keyword Args:
            title: Full Spoken Name.
            descr: Documentation Description.
            comment: Source Code Comment.
            ifdef: IFDEF pragma. Obsolete.
            ifdefs: IFDEFs pragmas.
            clkrel: Clock Relation (For signals and ports only).

        :meta public:
        """
        if self._is_locked:
            raise LockError(f"{self}: Cannot add item {name!r}.")
        ifdefs = cast_ifdefs(ifdefs or ifdef)
        items = self._items
        if name not in items.keys():
            doc = doc_from_type(type_, title=title, descr=descr, comment=comment)
            structitem = StructItem(name, type_, orientation=orientation, doc=doc, ifdefs=ifdefs, clkrel=clkrel)
            if not self.filter_ or self.filter_(structitem):
                items[name] = structitem
        else:
            raise ValueError(f"name {name!r} already exists in {self} ({self[name]})")

    def is_connectable(self, other) -> bool:
        """Check For Valid Connection To `other`."""
        if not isinstance(other, BaseStructType):
            return False
        try:
            return all(
                self._cmpitem(selfitem, otheritem)
                for selfitem, otheritem in zip(self.values(), other.values(), strict=True)
            )
        except ValueError:
            return False

    @staticmethod
    def _cmpitem(selfitem, otheritem):
        return (
            selfitem.name == otheritem.name
            and selfitem.type_.is_connectable(otheritem.type_)
            and selfitem.orientation == otheritem.orientation
            and selfitem.ifdefs == otheritem.ifdefs
        )

    @property
    def bits(self):
        """Size in Bits."""
        return sum(item.type_.bits for item in self.values())

    @abstractmethod
    def _build(self) -> None:
        """Build Type."""


class AStructType(BaseStructType, Light):
    """
    Base class for all structural types.

    The protected method `_build()` should be used to build the type.

    Definition of a struct:

    >>> import ucdp as u
    >>> class BusType(u.AStructType):
    ...     def _build(self) -> None:
    ...         self._add('data', u.UintType(8))
    ...         self._add('valid', u.BitType())
    ...         self._add('accept', u.BitType(), orientation=u.BWD)

    Usage of a Struct:

    >>> bus = BusType()
    >>> bus
    BusType()

    The structs behave like a `dict`, with elements hashed by `name`.
    But different to a regular `dict`, it returns items on pure iteration:

    >>> bus.keys()
    dict_keys(['data', 'valid', 'accept'])
    >>> bus.values()
    dict_values([StructItem('data', UintType(8)), StructItem('valid', BitType()), ...])
    >>> bus.items()
    dict_items([('data', StructItem('data', UintType(8))), ('valid', StructItem(...])
    >>> bus['valid']
    StructItem('valid', BitType())
    >>> bus.bits
    10

    Connections are only allowed to other :any:`AStructType` with the same key-value mapping.

    >>> BusType().is_connectable(BusType())
    True

    >>> class Bus2Type(u.AStructType):
    ...     def _build(self) -> None:
    ...         self._add('data', u.UintType(8))
    ...         self._add('valid', u.BitType())
    ...         self._add('accept', u.BitType(), orientation=u.FWD)
    >>> BusType().is_connectable(Bus2Type())
    False

    >>> class Bus3Type(u.AStructType):
    ...     def _build(self) -> None:
    ...         self._add('data', u.UintType(8), title="title")
    ...         self._add('valid', u.BitType(default=1))
    ...         self._add('accept', u.BitType(), orientation=u.BWD)
    >>> BusType().is_connectable(Bus3Type())
    True

    Struct members can be filtered.
    A struct member is added if the filter function returns `True` on a given :any:`StructItem` as argument.

    >>> def myfilter(structitem):
    ...     return "t" in structitem.name
    >>> for item in BusType(filter_=myfilter).values(): item
    StructItem('data', UintType(8))
    StructItem('accept', BitType(), orientation=BWD)

    There are also these predefined filters:

    >>> for item in BusType(filter_=u.fwdfilter).values(): item
    StructItem('data', UintType(8))
    StructItem('valid', BitType())

    >>> for item in BusType(filter_=u.bwdfilter).values(): item
    StructItem('accept', BitType(), orientation=BWD)

    This works also with the `new()` method:

    >>> for item in BusType().new(filter_=u.fwdfilter).values(): item
    StructItem('data', UintType(8))
    StructItem('valid', BitType())
    """

    def model_post_init(self, __context: Any) -> None:
        """Run Build."""
        self._build()
        self._is_locked = True


class AGlobalStructType(BaseStructType, Light):
    """
    A singleton struct which can be filled outside `_build` and is **shared** between instances.

    >>> import ucdp as u
    >>> class BusType(u.AGlobalStructType):
    ...     pass
    >>> bus = BusType()
    >>> bus.add('data', u.UintType(8))
    >>> bus.add('valid', u.BitType())
    >>> bus = BusType()
    >>> bus.add('accept', u.BitType(), orientation=u.BWD)
    >>> tuple(bus)
    ('data', 'valid', 'accept')

    This is forbidden on normal struct:

    >>> class BusType(u.AStructType):
    ...     def _build(self) -> None:
    ...         pass
    >>> bus = BusType()
    >>> bus._add('data', u.UintType(8))
    Traceback (most recent call last):
      ...
    ucdp.exceptions.LockError: BusType(): Cannot add item 'data'.
    """

    def add(
        self,
        name: str,
        type_: BaseType,
        orientation: Orientation = FWD,
        title: str | None = None,
        descr: str | None = None,
        comment: str | None = None,
        ifdef: str | None = None,
        ifdefs: Ifdefs | str | None = None,
    ) -> None:
        """
        Add member `name` with `type_` and `orientation`.

        Args:
            name: Name
            type_: Type
            orientation: Orientation

        Keyword Args:
            title: Full Spoken Name.
            descr: Documentation Description.
            comment: Source Code Comment.
            ifdef: IFDEF pragma. Obsolete.
            ifdefs: IFDEFs pragmas.
        """
        ifdefs = cast_ifdefs(ifdefs or ifdef)
        self._add(
            name,
            type_,
            orientation=orientation,
            title=title,
            descr=descr,
            comment=comment,
            ifdefs=ifdefs,
        )

    def _build(self) -> None:
        """Build Type."""

    def model_post_init(self, __context: Any) -> None:
        """Run Build."""
        if self.__class__ is AGlobalStructType:
            raise TypeError("Can't instantiate abstract class AGlobalStructType. Please create a subclass.")
        self._build()


class DynamicStructType(BaseStructType):
    """
    A singleton struct which can be filled outside `_build` and is **not** shared between instances.

    >>> import ucdp as u
    >>> class BusType(u.DynamicStructType):
    ...     pass
    >>> bus = BusType()
    >>> bus.add('data', u.UintType(8))
    >>> bus.add('valid', u.BitType())
    >>> tuple(bus)
    ('data', 'valid')
    >>> bus = BusType()
    >>> bus.add('accept', u.BitType(), orientation=u.BWD)
    >>> tuple(bus)
    ('accept',)

    This is forbidden on normal struct:

    >>> class BusType(u.AStructType):
    ...     def _build(self) -> None:
    ...         pass
    >>> bus = BusType()
    >>> bus._add('data', u.UintType(8))
    Traceback (most recent call last):
      ...
    ucdp.exceptions.LockError: BusType(): Cannot add item 'data'.
    """

    def add(
        self,
        name: str,
        type_: BaseType,
        orientation: Orientation = FWD,
        title: str | None = None,
        descr: str | None = None,
        comment: str | None = None,
        ifdef: str | None = None,
        ifdefs: Ifdefs | str | None = None,
    ) -> None:
        """
        Add member `name` with `type_` and `orientation`.

        Args:
            name: Name
            type_: Type
            orientation: Orientation

        Keyword Args:
            title: Full Spoken Name.
            descr: Documentation Description.
            comment: Source Code Comment.
            ifdef: IFDEF pragma. Obsolete.
            ifdefs: IFDEFs pragmas.
        """
        ifdefs = cast_ifdefs(ifdefs or ifdef)
        self._add(
            name,
            type_,
            orientation=orientation,
            title=title,
            descr=descr,
            comment=comment,
            ifdefs=ifdefs,
        )

    def _build(self) -> None:
        """Build Type."""

    def model_post_init(self, __context: Any) -> None:
        """Run Build."""
        self._build()


def fwdfilter(structitem):
    """Filter For Forward Elements In Structs."""
    return structitem.orientation == FWD


def bwdfilter(structitem):
    """Filter For Backward Elements In Structs."""
    return structitem.orientation == BWD



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Scalar Types.

* :any:`IntegerType`
* :any:`BitType`
* :any:`BoolType`
* :any:`RailType`
* :any:`UintType`
* :any:`SintType`
"""

from typing import Any, ClassVar

from humannum import hex_

from .casting import Casting
from .slices import DOWN, Slice
from .typebase import AScalarType, AVecType, BaseType
from .typebaseenum import BaseEnumType

INTEGER_WIDTH: int = 32


class IntegerType(AScalarType):
    """
    Native Signed 32-Bit Integer.

    Attributes:
        default: Default Value. 0 by default.

    The width is fixed to 32.

    Documentation defaults are empty by default:

    >>> import ucdp as u
    >>> u.IntegerType().title
    >>> u.IntegerType().descr
    >>> u.IntegerType().comment

    Example:
    >>> example = u.IntegerType()
    >>> example
    IntegerType()
    >>> example.width
    32
    >>> example.default
    0

    Another Example:

    >>> example = u.IntegerType(default=8)
    >>> example
    IntegerType(default=8)
    >>> example.width
    32
    >>> example.default
    8

    Selective Coverage Disable:

    >>> example = u.IntegerType()
    >>> example
    IntegerType()

    Range checking:

    >>> 5 in IntegerType()
    True
    >>> range(3, 10) in IntegerType()
    True
    >>> 2**32 in IntegerType()
    False

    Slicing:

    >>> u.IntegerType(default=31)['31:0']
    IntegerType(default=31)
    >>> u.IntegerType(default=31)['3:1']
    UintType(3, default=7)
    >>> u.IntegerType(default=31)['32:31']
    Traceback (most recent call last):
        ...
    ValueError: Cannot slice bit(s) 32:31 from IntegerType(default=31) with dimension [31:0]
    """

    width: ClassVar[int] = INTEGER_WIDTH  # type: ignore[misc]
    """Width in Bits."""

    min_: ClassVar[int] = -1 * 2 ** (width - 1)
    """Minimal Value."""

    max_: ClassVar[int] = 2 ** (width - 1) - 1
    """Maximal Value."""

    logic: bool = True
    """Include X and Z states, not just numeric values."""

    default: Any = 0
    """Default Value."""

    def check(self, value, what="Value") -> int:
        """
        Check `value` for type.

        Values are limited to 32-bit signed [-2147483648, 2147483647].

        >>> import ucdp as u
        >>> example = u.IntegerType()
        >>> example.check(0)
        0
        >>> example.check(2147483647)
        2147483647
        >>> example.check(2147483648)
        Traceback (most recent call last):
          ...
        ValueError: Value 2147483648 is not a 32-bit signed integer with range [-2147483648, 2147483647]
        >>> example.check(-2147483648)
        -2147483648
        >>> example.check(-2147483649)
        Traceback (most recent call last):
          ...
        ValueError: Value -2147483649 is not a 32-bit signed integer with range [-2147483648, 2147483647]
        """
        value = int(value)
        if (value < self.min_) or (value > self.max_):
            raise ValueError(f"{what} {value} is not a 32-bit signed integer with range [{self.min_}, {self.max_}]")
        return value

    def get_hex(self, value=None):
        """
        Return Hex Value.

        >>> import ucdp as u
        >>> u.IntegerType(default=0x10FEC0DE).get_hex()
        Hex('0x10FEC0DE')
        >>> u.IntegerType(default=0x10FEC0DE).get_hex(value=0xBEEF)
        Hex('0x0000BEEF')
        >>> u.IntegerType().get_hex(value=9)
        Hex('0x00000009')
        """
        if value is None:
            value = self.default
        self.check(value)
        value = int(value)
        wrap = 1 << 32
        value = (value + wrap) % wrap
        return hex_(value, width=32)

    def is_connectable(self, other) -> bool:
        """
        Check For Valid Connection To `other`.

        Connections are only allowed to other :any:`IntegerType`.

        >>> IntegerType().is_connectable(IntegerType())
        True
        >>> IntegerType(default=1).is_connectable(IntegerType(default=2))
        True
        """
        return (
            isinstance(other, (SintType, IntegerType))
            and int(other.width) == INTEGER_WIDTH  # type: ignore[operator]
            and self.logic == other.logic
        )

    def __getitem__(self, slice_):
        """Return Sliced Variant."""
        slice_ = Slice.cast(slice_, direction=DOWN)
        if slice_.width == INTEGER_WIDTH and slice_.right == 0:
            return self
        if slice_.left < (INTEGER_WIDTH - 1):
            return UintType(slice_.width, default=slice_.extract(self.default))
        raise ValueError(f"Cannot slice bit(s) {slice_!s} from {self} with dimension [31:0]")


class BitType(AScalarType):
    """
    Native Single Bit.

    Attributes:
        default: Default Value. 0 by default.

    The width is fixed to 1.

    Example:
    >>> import ucdp as u
    >>> example = u.BitType()
    >>> example
    BitType()
    >>> example.width
    1
    >>> example.default
    0

    Another Example:

    >>> example = u.BitType(default=1)
    >>> example
    BitType(default=1)
    >>> example.width
    1
    >>> example.default
    1

    Slicing:

    >>> u.BitType(default=1)[0]
    BitType(default=1)
    >>> u.BitType(default=1)[32:31]
    Traceback (most recent call last):
        ...
    ValueError: Cannot slice bit(s) 32:31 from BitType(default=1)
    """

    width: ClassVar[int] = 1  # type: ignore[misc]
    """Width in Bits."""

    min_: ClassVar[int] = 0  # type: ignore[misc]
    """Minimal Value."""

    max_: ClassVar[int] = 1  # type: ignore[misc]
    """Maximal Value."""

    logic: bool = True
    """Include X and Z states, not just numeric values."""

    default: Any = 0
    """Default Value."""

    @staticmethod
    def check(value, what="Value") -> int:
        """
        Check `value` for type.

        Values are limited to 0 and 1

        >>> import ucdp as u
        >>> example = u.BitType()
        >>> example.check(-1)
        Traceback (most recent call last):
          ...
        ValueError: Value -1 is not a single bit with range [0, 1]
        >>> example.check(0)
        0
        >>> example.check(1)
        1
        >>> example.check(2)
        Traceback (most recent call last):
          ...
        ValueError: Value 2 is not a single bit with range [0, 1]
        >>> example.check(False)
        0
        >>> example.check(True)
        1
        """
        value = int(value)
        if value not in (0, 1):
            raise ValueError(f"{what} {value} is not a single bit with range [0, 1]")
        return value

    def get_hex(self, value=None):
        """
        Return Hex Value.

        >>> import ucdp as u
        >>> u.BitType().get_hex()
        Hex('0x0')
        >>> u.BitType(default=1).get_hex()
        Hex('0x1')
        >>> u.BitType().get_hex(value=1)
        Hex('0x1')
        """
        if value is None:
            value = self.default
        self.check(value)
        return hex_(value, width=1)

    def is_connectable(self, other) -> bool:
        """
        Check For Valid Connection To `other`.

        Connections are only allowed to other :any:`BitType`.

        >>> BitType().is_connectable(BitType())
        True
        >>> BitType(default=1).is_connectable(BitType(default=0))
        True

        A connection to an :any:`UintType()` of width is forbidden (requires a cast).

        >>> BitType().is_connectable(UintType(2))
        False
        """
        return isinstance(other, BitType) and int(other.width) == 1 and self.logic == other.logic  # type: ignore[operator]

    def cast(self, other: BaseType) -> Casting:
        """
        How to cast an input of type `self` from a value of type `other`.

        `self = cast(other)`
        """
        if isinstance(other, (UintType, BitType, SintType, IntegerType, RailType)) and self.width == other.width:  # type: ignore[operator]
            return [("", "")]

        if isinstance(other, BaseEnumType) and self.width == other.keytype.width:  # type: ignore[operator]
            return [("", "")]

        return None

    def __getitem__(self, slice_):
        slice_ = Slice.cast(slice_, direction=DOWN)
        if slice_.width == 1 and slice_.right == 0:
            return self
        raise ValueError(f"Cannot slice bit(s) {slice_!s} from {self}")


class BoolType(AScalarType):
    """
    Native Boolean.

    Attributes:
        default: Default Value. 0 by default.

    The width is fixed to 1.

    Example:
    >>> import ucdp as u
    >>> example = u.BoolType()
    >>> example
    BoolType()
    >>> example.width
    1
    >>> example.default
    0

    Another Example:

    >>> example = u.BoolType(default=True)
    >>> example
    BoolType(default=True)
    >>> example.width
    1
    >>> example.default
    True

    Slicing:

    >>> u.BoolType(default=True)[0]
    BoolType(default=True)
    >>> u.BoolType()[32:31]
    Traceback (most recent call last):
        ...
    ValueError: Cannot slice bit(s) 32:31 from BoolType()
    """

    default: Any = False
    """Default Value."""

    width: ClassVar[int] = 1  # type: ignore[misc]
    """Width in Bits."""

    @staticmethod
    def check(value, what="Value") -> bool:
        """
        Check `value` for type.

        Values are limited to 0 and 1

        >>> import ucdp as u
        >>> example = u.BoolType()
        >>> example.check(-1)
        Traceback (most recent call last):
          ...
        ValueError: Value -1 is not a boolean
        >>> example.check(0)
        0
        >>> example.check(1)
        1
        >>> example.check(2)
        Traceback (most recent call last):
          ...
        ValueError: Value 2 is not a boolean
        >>> example.check(False)
        0
        >>> example.check(True)
        1
        """
        # Note: we need identity check here
        if int(value) in (0, 1):
            return value
        raise ValueError(f"{what} {value} is not a boolean")

    def get_hex(self, value=None):
        """
        Return Hex Value.

        >>> import ucdp as u
        >>> u.BoolType().get_hex()
        Hex('0x0')
        >>> u.BoolType(default=True).get_hex()
        Hex('0x1')
        >>> u.BoolType().get_hex(value=True)
        Hex('0x1')
        """
        if value is None:
            value = self.default
        self.check(value)
        return hex_(value, width=1)

    def is_connectable(self, other) -> bool:
        """
        Check For Valid Connection To `other`.

        Connections are only allowed to other :any:`BoolType`.

        >>> BoolType().is_connectable(BoolType())
        True
        >>> BoolType(default=True).is_connectable(BoolType(default=False))
        True

        A connection to an :any:`UintType()` is forbidden (requires a cast).

        >>> BoolType().is_connectable(UintType(1))
        False
        """
        return isinstance(other, BoolType)

    def __getitem__(self, slice_):
        slice_ = Slice.cast(slice_, direction=DOWN)
        if slice_.width == 1 and slice_.right == 0:
            return self
        raise ValueError(f"Cannot slice bit(s) {slice_!s} from {self}")


class RailType(AScalarType):
    """
    Voltage Rail.

    Attributes:
        default: Default Value. 0 by default.

    The width is fixed to 1.

    Example:
    >>> import ucdp as u
    >>> example = u.RailType()
    >>> example
    RailType()
    >>> example.width
    1
    >>> example.default

    Another Example:

    >>> example = u.RailType(default=1)
    >>> example
    RailType(default=1)
    >>> example.width
    1
    >>> example.default
    1

    Slicing:

    >>> u.RailType(default=1)[0]
    RailType(default=1)
    >>> u.RailType(default=1)[32:31]
    Traceback (most recent call last):
        ...
    ValueError: Cannot slice bit(s) 32:31 from RailType(default=1)
    """

    width: ClassVar[int] = 1  # type: ignore[misc]
    """Width in Bits."""

    logic: bool = True
    """Include X and Z states, not just numeric values."""

    default: Any | None = None
    """Default Value."""

    @staticmethod
    def check(value, what="Value") -> int:
        """
        Check `value` for type.

        Values are limited to 0 and 1

        >>> import ucdp as u
        >>> example = u.RailType()
        >>> example.check(-1)
        Traceback (most recent call last):
          ...
        ValueError: Value -1 is not a single bit with range [0, 1]
        >>> example.check(0)
        0
        >>> example.check(1)
        1
        >>> example.check(2)
        Traceback (most recent call last):
          ...
        ValueError: Value 2 is not a single bit with range [0, 1]
        >>> example.check(False)
        0
        >>> example.check(True)
        1
        """
        value = int(value)
        if value not in (0, 1):
            raise ValueError(f"{what} {value} is not a single bit with range [0, 1]")
        return value

    def get_hex(self, value=None):
        """
        Return Hex Value.

        >>> import ucdp as u
        >>> u.RailType().get_hex()
        >>> u.RailType(default=0).get_hex()
        Hex('0x0')
        >>> u.RailType(default=1).get_hex()
        Hex('0x1')
        >>> u.RailType().get_hex(value=1)
        Hex('0x1')
        """
        if value is None:
            value = self.default
        if value is None:
            return None
        self.check(value)
        return hex_(value, width=1)

    def is_connectable(self, other) -> bool:
        """
        Check For Valid Connection To `other`.

        Connections are only allowed to other :any:`RailType`.

        >>> RailType().is_connectable(RailType())
        True
        >>> RailType(default=1).is_connectable(RailType(default=0))
        True

        A connection to an :any:`BitType()` is forbidden (requires a cast).

        >>> RailType().is_connectable(BitType())
        False
        """
        return isinstance(other, RailType) and self.logic == other.logic

    def cast(self, other: BaseType) -> Casting:
        """
        How to cast an input of type `self` from a value of type `other`.

        `self = cast(other)`
        """
        if isinstance(other, (UintType, BitType, SintType, IntegerType, RailType)) and self.width == other.width:  # type: ignore[operator]
            return [("", "")]

        if isinstance(other, BaseEnumType) and self.width == other.keytype.width:  # type: ignore[operator]
            return [("", "")]

        return None

    def __getitem__(self, slice_):
        slice_ = Slice.cast(slice_, direction=DOWN)
        if slice_.width == 1 and slice_.right == 0:
            return self
        raise ValueError(f"Cannot slice bit(s) {slice_!s} from {self}")


class UintType(AVecType):
    """
    Vector With Unsigned Interpretation.

    Args:
        width (int): Width in bits.

    Attributes:
        default: Default Value. 0 by default.

    Example:
    >>> import ucdp as u
    >>> example = u.UintType(12)
    >>> example
    UintType(12)
    >>> example.width
    12
    >>> example.default
    0

    Another Example:

    >>> example = u.UintType(16, default=8)
    >>> example
    UintType(16, default=8)
    >>> example.width
    16
    >>> example.default
    8

    Selective Coverage Disable:

    >>> example = u.UintType(32)
    >>> example
    UintType(32)

    Slicing:

    >>> u.UintType(16, default=31)[15:0]
    UintType(16, default=31)
    >>> u.UintType(16, default=31)[3:1]
    UintType(3, default=7)
    >>> u.UintType(16, default=31)[16:15]
    Traceback (most recent call last):
        ...
    ValueError: Cannot slice bit(s) 16:15 from UintType(16, default=31)
    """

    min_: ClassVar[int] = 0
    """Minimal Value."""

    default: Any = 0
    """Default Value."""

    def __init__(self, width, **kwargs):
        super().__init__(width=width, **kwargs)

    @property
    def max_(self):
        """Maximum Value."""
        return (2**self.width) - 1

    def check(self, value, what="Value") -> int:
        """
        Check `value` for type.

        >>> import ucdp as u
        >>> example = u.UintType(8)
        >>> example.check(0)
        0
        >>> example.check(255)
        255
        >>> example.check(256)
        Traceback (most recent call last):
          ...
        ValueError: Value 256 is not a 8-bit integer with range [0, 255]
        >>> example.check(-1)
        Traceback (most recent call last):
          ...
        ValueError: Value -1 is not a 8-bit integer with range [0, 255]
        """
        value = int(value)

        if (value < 0) or (value > int(self.max_)):
            raise ValueError(f"{what} {value} is not a {self.width}-bit integer with range [0, {self.max_}]")
        return value

    def get_hex(self, value=None):
        """
        Return Hex Value.

        >>> import ucdp as u
        >>> u.UintType(9).get_hex()
        Hex('0x000')
        >>> u.UintType(9, default=0xFE).get_hex()
        Hex('0x0FE')
        >>> u.UintType(9).get_hex(value=0xFE)
        Hex('0x0FE')
        """
        if value is None:
            value = self.default
        self.check(value)
        return hex_(value, width=self.width)

    def is_connectable(self, other) -> bool:
        """
        Check For Valid Connection To `other`.

        Connections are only allowed to other :any:`UintType` of the same width.

        >>> UintType(8).is_connectable(UintType(8))
        True
        >>> UintType(8).is_connectable(UintType(9))
        False
        >>> UintType(8, default=1).is_connectable(UintType(8, default=0))
        True

        A connection to an :any:`BitType()` is allowed, but :any:`SintType()` is forbidden (requires a cast).

        >>> UintType(1).is_connectable(BitType())
        True
        >>> UintType(1).is_connectable(SintType(1))
        False
        """
        return isinstance(other, (UintType, BitType)) and self.width == other.width and self.logic == other.logic  # type: ignore[operator]

    def cast(self, other: BaseType) -> Casting:
        """
        How to cast an input of type `self` from a value of type `other`.

        `self = cast(other)`
        """
        if isinstance(other, (SintType, IntegerType)) and self.width == other.width:  # type: ignore[operator]
            return [("", "")]

        if isinstance(other, BaseEnumType) and self.width == other.keytype.width:
            return [("", "")]

        return None

    def __getitem__(self, slice_):
        slice_ = Slice.cast(slice_, direction=DOWN)
        if slice_.width == self.width and slice_.right == self.right:
            return self
        if slice_ in self.slice_:
            return UintType(slice_.width, default=slice_.extract(self.default))
        raise ValueError(f"Cannot slice bit(s) {slice_!s} from {self}")


class SintType(AVecType):
    """
    Vector With Unsigned Interpretation.

    Args:
        width (int): Width in bits.

    Attributes:
        default: Default Value. 0 by default.

    Example:
    >>> import ucdp as u
    >>> example = u.SintType(12)
    >>> example
    SintType(12)
    >>> example.width
    12
    >>> example.default
    0

    Another Example:

    >>> example = u.SintType(16, default=8)
    >>> example
    SintType(16, default=8)
    >>> example.width
    16
    >>> example.default
    8

    Slicing:

    >>> u.SintType(16, default=31)[15:0]
    SintType(16, default=31)
    >>> u.SintType(16, default=31)[15:8]
    SintType(8)
    >>> u.SintType(16, default=31)[3:1]
    UintType(3, default=7)
    >>> u.SintType(16, default=31)[16:15]
    Traceback (most recent call last):
        ...
    ValueError: Cannot slice bit(s) 16:15 from SintType(16, default=31)
    """

    default: Any = 0
    """Default Value."""

    @property
    def min_(self):
        """Minimal Value."""
        return -1 * 2 ** (self.width - 1)

    @property
    def max_(self):
        """Maximum Value."""
        return 2 ** (self.width - 1) - 1

    def check(self, value, what="Value") -> int:
        """
        Check `value` for type.

        >>> import ucdp as u
        >>> example = u.SintType(8)
        >>> example.check(-128)
        -128
        >>> example.check(0)
        0
        >>> example.check(127)
        127
        >>> example.check(128)
        Traceback (most recent call last):
          ...
        ValueError: Value 128 is not a 8-bit signed integer with range [-128, 127]
        >>> example.check(-129)
        Traceback (most recent call last):
          ...
        ValueError: Value -129 is not a 8-bit signed integer with range [-128, 127]
        """
        # if width < 1:
        #     raise ValueError(f"Invalid width {width}")
        value = int(value)
        if (value < int(self.min_)) or (value > int(self.max_)):
            raise ValueError(
                f"{what} {value} is not a {self.width}-bit signed integer with range [{self.min_}, {self.max_}]"
            )
        return value

    def get_hex(self, value=None):
        """
        Return Hex Value.

        >>> import ucdp as u
        >>> u.SintType(9).get_hex()
        Hex('0x000')
        >>> u.SintType(9, default=0xFE).get_hex()
        Hex('0x0FE')
        >>> u.SintType(9, default=-20).get_hex()
        Hex('0x1EC')
        >>> u.SintType(9).get_hex(value=0xFE)
        Hex('0x0FE')
        """
        if value is None:
            value = self.default
        self.check(value)
        width = int(self.width)
        wrap = 1 << width
        value = (value + wrap) % wrap
        return hex_(value, width=width)

    def is_connectable(self, other) -> bool:
        """
        Check For Valid Connection To `other`.

        Connections are only allowed to other :any:`SintType` of the same width.

        >>> SintType(8).is_connectable(SintType(8))
        True
        >>> SintType(8).is_connectable(SintType(9))
        False
        >>> SintType(8, default=1).is_connectable(SintType(8, default=0))
        True

        A connection to an :any:`BitType()` or :any:`UintType()` is forbidden (requires a cast).

        >>> SintType(1).is_connectable(BitType())
        False
        >>> SintType(1).is_connectable(UintType(1))
        False
        """
        return isinstance(other, (SintType)) and self.width == other.width and self.logic == other.logic  # type: ignore[operator]

    def cast(self, other: BaseType) -> Casting:
        """
        How to cast an input of type `self` from a value of type `other`.

        `self = cast(other)`
        """
        if isinstance(other, (UintType, BitType, IntegerType)) and self.width == other.width:  # type: ignore[operator]
            return [("", "")]

        if isinstance(other, BaseEnumType) and self.width == other.keytype.width:
            return [("", "")]

        return None

    def __getitem__(self, slice_):
        slice_ = Slice.cast(slice_, direction=DOWN)
        if slice_.width == self.width and slice_.right == self.right:
            return self
        if slice_ in self.slice_:
            if slice_.left == self.slice_.left:
                return SintType(slice_.width, default=slice_.extract(self.default))
            return UintType(slice_.width, default=slice_.extract(self.default))
        raise ValueError(f"Cannot slice bit(s) {slice_!s} from {self}")



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Module File Information."""

import re
from logging import INFO, WARNING

from pytest import raises

import ucdp as u
from ucdp import IN, OUT, Assign, ConstExpr, Default, Note, Port, UintType


class SubMod(u.AMod):
    """Sub."""

    def _build(self):
        self.add_port(u.UintType(4), "in_i")
        self.add_port(u.UintType(4), "out_o")
        self.add_port(u.UintType(4), "open_i")
        self.add_port(u.UintType(4), "open_o")
        self.add_port(u.UintType(4), "note_i")
        self.add_port(u.UintType(4), "note_o")
        self.add_port(u.UintType(4), "default_i")
        self.add_port(u.UintType(4), "default_o")


class TopMod(u.AMod):
    """Top."""

    def _build(self):
        self.add_port(u.UintType(4), "in_i")
        self.add_port(u.UintType(4), "out_o")

        sub = SubMod(self, "u_sub0")
        sub.con("in_i", "4'h4")
        sub.con("out_o", "out_o")
        sub.con("open_i", u.OPEN)
        sub.con("open_o", u.OPEN)
        sub.con("note_i", u.note("my note"))
        sub.con("note_o", u.note("other note"))
        sub.con("default_i", u.DEFAULT)
        sub.con("default_o", u.DEFAULT)


def test_top():
    """Top Module."""
    top = TopMod()
    assert tuple(top.get_instcons("u_sub0").iter()) == (
        Assign(target=Port(UintType(4), "in_i", direction=IN), source=ConstExpr(UintType(4, default=4))),
        Assign(target=Port(UintType(4), "out_o", direction=OUT), source=Port(UintType(4), "out_o", direction=OUT)),
        Assign(target=Port(UintType(4), "open_i", direction=IN), source=Note(note="OPEN")),
        Assign(target=Port(UintType(4), "open_o", direction=OUT), source=Note(note="OPEN")),
        Assign(target=Port(UintType(4), "note_i", direction=IN), source=Note(note="my note")),
        Assign(target=Port(UintType(4), "note_o", direction=OUT), source=Note(note="other note")),
        Assign(target=Port(UintType(4), "default_i", direction=IN), source=Default(note="DEFAULT")),
        Assign(target=Port(UintType(4), "default_o", direction=OUT), source=Default(note="DEFAULT")),
    )


class TopErrMod(u.AMod):
    """Top Module with Routing Error."""

    def _build(self):
        self.add_port(u.UintType(4), "in_i")
        self.add_port(u.UintType(4), "out_o")

        sub = SubMod(self, "u_sub0")
        sub.con("in_i", "5'h4")
        sub.con("out_o", "out_o")


def test_top_err(caplog):
    """Top Module with Routing Error."""
    msg = (
        "<tests.test_mod_con.TopErrMod(inst='top_err', libname='tests', modname='top_err')>: "
        "Cannot assign 'ConstExpr(UintType(5, default=4))' of type UintType(5, default=4) to 'in_i' "
        "of type UintType(4)."
    )
    with raises(TypeError, match=re.escape(msg)):
        TopErrMod()


class TopWarnMod(u.AMod):
    """Top Module with Routing Warning."""

    def _build(self):
        self.add_port(u.UintType(4), "in_i")
        self.add_port(u.UintType(4), "out_o")

        sub = SubMod(self, "u_sub0")
        sub.con("in_i", "5'h4", on_error="warn")
        sub.con("out_o", "out_o")


def test_top_warn(caplog):
    """Top Module with Routing Error."""
    msg = (
        "<tests.test_mod_con.TopWarnMod(inst='top_warn', libname='tests', modname='top_warn')>: "
        "Cannot assign 'ConstExpr(UintType(5, default=4))' of type UintType(5, default=4) to 'in_i' "
        "of type UintType(4)."
    )
    top = TopWarnMod()
    assert tuple(top.get_instcons("u_sub0").iter()) == (
        Assign(target=Port(UintType(4), "in_i", direction=IN)),
        Assign(target=Port(UintType(4), "out_o", direction=OUT), source=Port(UintType(4), "out_o", direction=OUT)),
        Assign(target=Port(UintType(4), "open_i", direction=IN)),
        Assign(target=Port(UintType(4), "open_o", direction=OUT)),
        Assign(target=Port(UintType(4), "note_i", direction=IN)),
        Assign(target=Port(UintType(4), "note_o", direction=OUT)),
        Assign(target=Port(UintType(4), "default_i", direction=IN)),
        Assign(target=Port(UintType(4), "default_o", direction=OUT)),
    )
    assert caplog.record_tuples == [("ucdp", WARNING, msg)]


class TopIgnoreMod(u.AMod):
    """Top Module with ignored Routing Error."""

    def _build(self):
        self.add_port(u.UintType(4), "in_i")
        self.add_port(u.UintType(4), "out_o")

        sub = SubMod(self, "u_sub0")
        sub.con("in_i", "5'h4", on_error="ignore")
        sub.con("out_o", "out_o")


def test_top_ign(caplog):
    """Top Module with ignored Routing Error."""
    msg = (
        "Ignored: <tests.test_mod_con.TopIgnoreMod(inst='top_ignore', libname='tests', modname='top_ignore')>: "
        "Cannot assign 'ConstExpr(UintType(5, default=4))' of type UintType(5, default=4) to 'in_i' "
        "of type UintType(4)."
    )
    top = TopIgnoreMod()
    assert tuple(top.get_instcons("u_sub0").iter()) == (
        Assign(target=Port(UintType(4), "in_i", direction=IN)),
        Assign(target=Port(UintType(4), "out_o", direction=OUT), source=Port(UintType(4), "out_o", direction=OUT)),
        Assign(target=Port(UintType(4), "open_i", direction=IN)),
        Assign(target=Port(UintType(4), "open_o", direction=OUT)),
        Assign(target=Port(UintType(4), "note_i", direction=IN)),
        Assign(target=Port(UintType(4), "note_o", direction=OUT)),
        Assign(target=Port(UintType(4), "default_i", direction=IN)),
        Assign(target=Port(UintType(4), "default_o", direction=OUT)),
    )
    assert caplog.record_tuples == [("ucdp", INFO, msg)]



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Descriptive Struct Type.

This module serves a struct type variant :any:`DescriptiveStructType` which describes a given `type_` with constants.
"""

from .nameutil import str2identifier
from .orientation import BWD, FWD
from .typearray import ArrayType
from .typebase import BaseType
from .typeenum import BaseEnumType
from .typescalar import IntegerType, SintType, UintType
from .typestruct import AStructType, BaseStructType, StructFilter


class DescriptiveStructType(AStructType):
    """
    Struct with constants describing `type_`.

    Attributes:
        type_: Type to be described

    Bit:

        >>> import ucdp as u
        >>> for item in u.DescriptiveStructType(u.BitType(default=1)).values():
        ...     repr(item)
        "StructItem('default_p', BitType(default=1), doc=Doc(title='Default Value'))"

    Unsigned Integer:

        >>> for item in u.DescriptiveStructType(u.UintType(16, default=3),
        ...                                   filter_=lambda item: item.name in ("default_p", "max_p")).values():
        ...     item
        StructItem('max_p', UintType(16, default=65535), doc=Doc(title='Maximal Value'))
        StructItem('default_p', UintType(16, default=3), doc=Doc(title='Default Value'))

    Signed Integer:

        >>> for item in u.DescriptiveStructType(u.SintType(8, default=-3)).values():
        ...     repr(item)
        "StructItem('width_p', IntegerType(default=8), doc=Doc(title='Width in Bits'))"
        "StructItem('min_p', SintType(8, default=-128), doc=Doc(title='Minimal Value'))"
        "StructItem('max_p', SintType(8, default=127), doc=Doc(title='Maximal Value'))"
        "StructItem('default_p', SintType(8, default=-3), doc=Doc(title='Default Value'))"

    Enum:

        >>> class MyEnumType(u.AEnumType):
        ...     keytype: u.AScalarType = u.UintType(2, default=0)
        ...     def _build(self) -> None:
        ...         self._add(0, "Forward")
        ...         self._add(1, 8)
        ...         self._add(2, "Bi-Dir")
        >>> for item in u.DescriptiveStructType(MyEnumType()).values():
        ...     repr(item)
        "StructItem('width_p', IntegerType(default=2), doc=Doc(title='Width in Bits'))"
        "StructItem('min_p', MyEnumType(default=0), doc=Doc(title='Minimal Value'))"
        "StructItem('max_p', MyEnumType(default=3), doc=Doc(title='Maximal Value'))"
        "StructItem('forward_e', UintType(2))"
        "StructItem('v8_e', UintType(2, default=1))"
        "StructItem('bi_dir_e', UintType(2, default=2))"
        "StructItem('default_p', MyEnumType(), doc=Doc(title='Default Value'))"

    Struct:

        >>> class SubStructType(u.AStructType):
        ...     def _build(self) -> None:
        ...         self._add('a', u.UintType(2))
        ...         self._add('b', u.UintType(3), u.BWD)
        >>> class MyStructType(u.AStructType):
        ...     def _build(self) -> None:
        ...         self._add('ctrl', u.UintType(4), title='Control')
        ...         self._add('data', u.ArrayType(u.SintType(16, default=5), 8), u.FWD, descr='Data to be handled')
        ...         self._add('resp', u.BitType(), u.BWD, comment='Sink response')
        ...         self._add('mode', MyEnumType(), u.BWD, comment='Enum')
        ...         self._add('sub', SubStructType(), u.BWD)
        >>> for item in u.DescriptiveStructType(MyStructType()).values():
        ...     repr(item)
        "StructItem('bits_p', IntegerType(default=140), doc=Doc(title='Size in Bits'))"
        "StructItem('fwdbits_p', IntegerType(default=135), doc=Doc(title='Forward Size in Bits'))"
        "StructItem('bwdbits_p', IntegerType(default=5), doc=Doc(title='Backward Size in Bits'))"
        "StructItem('bibits_p', IntegerType(), doc=Doc(title='Bi-Directional Size in Bits'))"

    """

    # All this works for parameterized types too:

    # >>> width_p = u.Param(u.IntegerType(default=16), 'width_p')
    # >>> type_ = u.UintType(width_p, default=4)
    # >>> for item in u.DescriptiveStructType(type_).values():
    # ...     repr(item)
    # "StructItem('width_p', IntegerType(default=Param(IntegerType(default=16), 'width_p')), doc=...
    # "StructItem('min_p', UintType(Param(IntegerType(default=16), 'width_p')), doc=...
    # "StructItem('max_p', UintType(Param(IntegerType(default=16), 'width_p'), default=Op(Op(2, '**', Param(...
    # "StructItem('default_p', UintType(Param(IntegerType(default=16), 'width_p'), default=4), doc=...

    type_: BaseType
    enumitem_suffix: str = "e"
    filter_: StructFilter | None = None

    def __init__(self, type_: BaseType, **kwargs):
        super().__init__(type_=type_, **kwargs)  # type: ignore[call-arg]

    def _build(self) -> None:
        self._add_type(self.type_)

    def _add_type(self, type_: BaseType) -> None:
        # Width
        if isinstance(type_, (UintType, SintType, BaseEnumType)):
            title = "Width in Bits"
            self._add("width_p", IntegerType(default=type_.width), title=title)
            title = "Minimal Value"
            self._add("min_p", type_.new(default=type_.min_), title=title)
            title = "Maximal Value"
            self._add("max_p", type_.new(default=type_.max_), title=title)
        if isinstance(type_, (ArrayType, BaseStructType)):
            title = "Size in Bits"
            self._add("bits_p", IntegerType(default=type_.bits), title=title)
            fwdbits, bwdbits, bibits = _get_dirwith(type_)
            title = "Forward Size in Bits"
            self._add("fwdbits_p", IntegerType(default=fwdbits), title=title)
            title = "Backward Size in Bits"
            self._add("bwdbits_p", IntegerType(default=bwdbits), title=title)
            title = "Bi-Directional Size in Bits"
            self._add("bibits_p", IntegerType(default=bibits), title=title)
        # Enum Items
        if isinstance(type_, BaseEnumType):
            enumitem_suffix = self.enumitem_suffix
            for eitem in type_.values():
                ename = str2identifier(str(eitem.value))
                etype = type_.keytype.new(default=eitem.key)
                doc = eitem.doc
                self._add(f"{ename}_{enumitem_suffix}", etype, title=doc.title, descr=doc.descr, comment=doc.comment)
        # Default
        if not isinstance(type_, (BaseStructType, ArrayType)):
            title = "Default Value"
            self._add("default_p", type_, title=title)


def _get_dirwith(type_, orientation=FWD):
    fwd, bwd, bid = 0, 0, 0
    for item in type_.values():
        itype_ = item.type_
        iorientation = item.orientation * orientation
        if isinstance(itype_, BaseStructType):
            ifwd, ibwd, ibid = _get_dirwith(itype_, iorientation)
            fwd += ifwd
            bwd += ibwd
            bid += ibid
        elif iorientation == FWD:
            fwd += itype_.bits
        elif iorientation == BWD:
            bwd += itype_.bits
        else:
            raise AssertionError
    return fwd, bwd, bid



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test :any:`Object`."""

from hypothesis import given
from hypothesis import strategies as st
from pydantic import ValidationError
from pytest import raises

import ucdp as u


class MyObject(u.Object):
    """Example Object."""

    arg1: int


class MyMyObject(MyObject):
    """Example Sub-Object."""

    arg2: int


class MyTuplingObject(u.Object):
    """Example Object with Tuple."""

    arg1: tuple[int, ...] = u.Field(default_factory=tuple)


class MyListingObject(u.Object):
    """Example Object with List."""

    arg1: list[int] = u.Field(default_factory=list)


@given(int1=st.integers(), int2=st.integers(), str1=st.text())
def test_object_basics(int1, int2, str1):
    """:any:`Object` Basic Testing."""
    inst = MyObject(arg1=int1)
    assert inst.arg1 == int1

    # compare
    assert MyObject(arg1=int1) == MyObject(arg1=int1)

    # no caching
    assert MyObject(arg1=int1) is not MyObject(arg1=int1)

    assert str(MyObject(arg1=int1)) == f"MyObject(arg1={int1!r})"
    assert repr(MyObject(arg1=int1)) == f"MyObject(arg1={int1!r})"

    # immutable
    with raises(ValidationError):
        inst.arg1 = int2

    # no extra
    with raises(ValidationError):
        MyObject(arg1=int1, arg2=int2)

    # validate
    with raises(ValidationError):
        MyObject(arg1=str1)

    assert hash(inst) is not None


@given(int1=st.integers(), int2=st.integers(), int3=st.integers(), str1=st.text())
def test_sub_object(int1, int2, int3, str1):
    """Sub :any:`Object` Testing."""
    inst = MyMyObject(arg1=int1, arg2=int2)
    assert inst.arg1 == int1
    assert inst.arg2 == int2

    with raises(ValidationError):
        inst.arg1 = int3

    with raises(ValidationError):
        MyMyObject(arg1=int1, arg2=str1)

    # no caching
    assert MyMyObject(arg1=int1, arg2=int2) is not MyMyObject(arg1=int1, arg2=int2)
    assert MyMyObject(arg1=int1, arg2=int2) is not MyMyObject(arg2=int2, arg1=int1)
    assert MyMyObject(arg1=int1, arg2=int2) is not MyMyObject(arg1=int1, arg2=int3)
    assert MyMyObject(arg1=int1, arg2=int3) is not MyMyObject(arg1=int1, arg2=int3)


@given(int1=st.integers(), int2=st.integers(), tuple1=st.tuples(st.integers()))
def test_tupling_object(int1, int2, tuple1):
    """:any:`Object` With Tuple Testing."""
    assert MyTuplingObject().arg1 == ()

    with raises(ValidationError):
        MyTuplingObject(arg1=int1)

    inst = MyTuplingObject(arg1=tuple1)
    assert inst.arg1 == tuple1

    assert hash(inst) is not None


@given(int1=st.integers(), int2=st.integers(), list1=st.lists(st.integers()))
def test_listing_object(int1, int2, list1):
    """:any:`Object` With List Testing."""
    assert MyListingObject().arg1 == []

    with raises(ValidationError):
        MyListingObject(arg1=int1)

    inst = MyListingObject(arg1=list1)
    assert inst.arg1 == list1

    inst.arg1.append(int2)

    assert inst.arg1[-1] == int2

    with raises(TypeError):
        hash(inst)


class OneObject(u.Object):
    """Example Object."""

    arg1: int


class ParentObject(u.Object):
    """Example Object."""

    arg1: int
    arg2: OneObject
    arg3: tuple[OneObject, ...]


@given(int1=st.integers(), int2=st.integers(), int3=st.integers(), int4=st.integers())
def test_nested(int1, int2, int3, int4):
    """Nested Types."""
    one2 = OneObject(arg1=int2)
    one3 = OneObject(arg1=int3)
    one4 = OneObject(arg1=int4)
    parent = ParentObject(arg1=int1, arg2=one2, arg3=(one3, one4))
    assert parent.arg1 == int1
    assert parent.arg2 == one2
    assert parent.arg3 == (one3, one4)



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Clock Relation.
"""

from .clkrelbase import BaseClkRel
from .humannum import Freq
from .signal import BaseSignal


class ClkRel(BaseClkRel):
    """
    Clock Relation.

    >>> import ucdp as u
    >>> clk = u.Port(type_=u.ClkType(), name="clk_i")
    >>> inp = u.Port(type_=u.UintType(8), name="inp_i", clkrel=u.ASYNC)
    >>> inp.clkrel
    ASYNC
    >>> inp.clkrel.info
    'ASYNC'
    >>> outp = u.Port(type_=u.UintType(8), name="outp_o", clkrel=u.ClkRel(clk=clk))
    >>> outp.clkrel
    ClkRel(clk=Port(ClkType(), 'clk_i', direction=IN))
    >>> outp.clkrel.info
    'clk: clk_i'

    """

    clk: BaseSignal | str | None = None
    """Signal."""

    freq_min: Freq | None = None
    freq: Freq | None = None
    freq_max: Freq | None = None

    @property
    def freqs(self) -> tuple[Freq | None, Freq | None, Freq | None]:
        """Frequency Tuple: Min/Typ/Max."""
        return (self.freq_min, self.freq, self.freq_max)

    @property
    def info(self) -> str:
        """Information."""
        result = []
        clk = self.clk
        if clk:
            if isinstance(clk, BaseSignal):
                clk = clk.name
            result.append(f"clk: {clk}")
        freqs = self.freqs
        if any(freqs):
            result.append(str(freqs))
        return " ".join(result)


class _AsyncClkRel(BaseClkRel):
    """Async."""

    @property
    def info(self) -> str:
        """Information."""
        return "ASYNC"

    def __repr__(self) -> str:
        return "ASYNC"

    def __str__(self) -> str:
        return "ASYNC"


ASYNC = _AsyncClkRel()



# MIT License
#
# Copyright (c) 2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Pytest Configuration and Fixtures."""

import os
from contextlib import contextmanager
from pathlib import Path
from unittest import mock

from pytest import fixture

import ucdp as u

EXAMPLES_PATH = Path(u.__file__).parent / "examples"
TESTDATA_PATH = Path(__file__).parent / "testdata"
TESTS_PATH = Path(__file__).parent


@contextmanager
def _env(prjroot: Path, path: Path | None = None):
    """Environment."""
    path = path or prjroot
    env = {
        "PRJROOT": str(prjroot),
        "UCDP_PATH": str(path),
        "UCDP_MAXWORKERS": "1",
        "UCDP_NO_COLOR": "1",
    }

    with u.extend_sys_path((path,)):
        with mock.patch.dict(os.environ, env):
            yield path


@fixture
def prjroot(tmp_path):
    """Add access to ``examples/simple``."""
    yield tmp_path


@fixture
def example_simple(prjroot):
    """Add access to ``examples/simple``."""
    with _env(prjroot, path=EXAMPLES_PATH / "simple") as path:
        yield path


@fixture
def example_bad(prjroot):
    """Add access to ``examples/bad``."""
    with _env(prjroot, path=EXAMPLES_PATH / "bad") as path:
        yield path


@fixture
def example_filelist(prjroot):
    """Add access to ``examples/filelist``."""
    with _env(prjroot, path=EXAMPLES_PATH / "filelist") as path:
        yield path


@fixture
def example_param(prjroot):
    """Add access to ``examples/param``."""
    with _env(prjroot, path=EXAMPLES_PATH / "param") as path:
        yield path


@fixture
def testdata(prjroot):
    """Add access to ``testdata``."""
    with _env(prjroot, path=TESTDATA_PATH) as path:
        yield path


@fixture
def tests(prjroot):
    """Add access to ``tests``."""
    with _env(prjroot, path=TESTS_PATH.parent) as path:
        yield path


@fixture
def examples_path(prjroot):
    """Add access to ``examples``."""
    yield EXAMPLES_PATH


@fixture
def uartcorefile(prjroot):
    """UART Core File."""
    uartcorefile = prjroot / "uart_lib" / "uart" / "rtl" / "uart_core.sv"
    uartcorefile.parent.mkdir(parents=True, exist_ok=True)
    uartcorefile.touch()
    yield



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Command Line Interface - Utilities."""

from collections.abc import Iterator
from logging import FATAL, getLogger
from pathlib import Path
from typing import IO, Any

import click
from rich.console import Console

from ucdp.finder import find
from ucdp.pathutil import improved_glob

PathType = click.Path(path_type=Path)


def auto_top(ctx, param, incomplete):
    """Autocompletion for TOP."""
    getLogger().setLevel(FATAL)
    infos = find(patterns=[f"{incomplete}*"], glob=True)
    return [str(info.topmodref) for info in infos]


def auto_path(ctr, param, incomplete):
    """Autocompletion for Paths."""
    return [str(path) for path in improved_glob(Path(f"{incomplete}*"))]


arg_top = click.argument("top", envvar="UCDP_TOP", shell_complete=auto_top)
arg_tops = click.argument("tops", nargs=-1, envvar="UCDP_TOP", shell_complete=auto_top)
opt_topsfile = click.option(
    "--tops-file",
    type=click.Path(path_type=Path),
    default=[],
    multiple=True,
    help="File with Top Module References",
)

opt_path = click.option(
    "--path",
    "-p",
    default=[],
    multiple=True,
    envvar="UCDP_PATH",
    type=click.Path(path_type=Path),
    shell_complete=auto_path,
    help="""
Search Path For Data Model And Template Files.
This option can be specified multiple times.
Environment Variable 'UCDP_PATH'.
""",
)
opt_filelist = click.option(
    "--filelist",
    "-f",
    multiple=True,
    help="Filelist Names. Environment Variable 'UCDP_FILELIST'.",
    envvar="UCDP_FILELIST",
)
opt_target = click.option(
    "--target",
    "-t",
    help="Filter File List for Target. Environment Variable 'UCDP_TARGET'.",
    envvar="UCDP_TARGET",
)
opt_show_diff = click.option(
    "--show-diff",
    "-s",
    default=False,
    is_flag=True,
    help="Show What Changed. Environment Variable 'UCDP_SHOW_DIFF'.",
    envvar="UCDP_SHOW_DIFF",
)
opt_maxlevel = click.option("--maxlevel", "-L", type=int, help="Limit to maximum number of hierarchy levels.")
opt_dry_run = click.option("--dry-run", default=False, is_flag=True, help="Do nothing.")
opt_maxworkers = click.option(
    "--maxworkers",
    "-J",
    type=int,
    help="Maximum Number of Processes.",
    envvar="UCDP_MAXWORKERS",
)
opt_defines = click.option(
    "--define",
    "-D",
    multiple=True,
    type=str,
    help="Defines set on the datamodel. Environment Variable 'UCDP_DEFINES'",
    envvar="UCDP_DEFINES",
)
opt_file = click.option(
    "--file",
    "-o",
    type=click.File("w"),
    shell_complete=auto_path,
    help="Output to file instead of STDOUT",
)
opt_filepath = click.option(
    "--file",
    "-o",
    type=click.Path(path_type=Path),
    shell_complete=auto_path,
    help="Output to file instead of STDOUT",
)
opt_tag = click.option(
    "--tag",
    "-G",
    default=[],
    multiple=True,
    help="Filter Modules by Tag Name or Wildcard.",
)
arg_template_filepaths = click.argument(
    "template_filepaths",
    type=PathType,
    shell_complete=auto_path,
    nargs=-1,
    envvar="UCDP_TEMPLATE_FILEPATHS",
)
opt_local = click.option(
    "--local/--no-local",
    "-l/-L",
    default=True,
    is_flag=True,
    help="List local/non-local modules only. Selected by default.",
)
opt_check = click.option(
    "--check",
    default=False,
    is_flag=True,
    help="Report an error if any file changes.",
)
opt_create = click.option(
    "--create",
    "-c",
    default=False,
    is_flag=True,
    help="Create missing inplace files.",
)
opt_clean = click.option(
    "--clean",
    "-n",
    default=False,
    is_flag=True,
    help="Remove obsolete fully-generated files.",
)


def defines2data(defines: list[str]) -> dict[str, str]:
    """Convert defines to data."""
    return dict(define.split("=", 1) if "=" in define else (define, None) for define in defines)


def read_file(filepath: Path) -> Iterator[str]:
    """Read File."""
    with filepath.open() as file:
        for line in file:
            line = line.strip()  # noqa: PLW2901
            if line.startswith("#"):
                continue
            if line:
                yield line


class Exit(click.ClickException):
    """Graceful Exit."""

    def __init__(self, console: Console, message: str) -> None:
        super().__init__(message)
        self.console = console

    def show(self, file: IO[Any] | None = None) -> None:
        """Show."""
        self.console.print(f"[red][bold]{self.message}.")



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Type Orientation and Port Directions.

Types can have an orientation (:any:`Orientation`) and ports have a :any:`Direction`.


Orient
===========

A member in a structural type can have an orientation:

* forward (`FWD`) from source to sink
* backward (`BWD`) from sink to source

>>> from tabulate import tabulate
>>> import ucdp as u
>>> orientations = (u.FWD, u.BWD)
>>> overview = [(d, d.mode, d.name, d.suffix) for d in orientations]
>>> print(tabulate(overview, ("Orient", ".mode", ".name", ".suffix")))
Orient      .mode  .name    .suffix
--------  -------  -------  ---------
FWD             1  FWD
BWD            -1  BWD

Direction
=========

Ports have a direction:

* input (`IN`)
* output (`OUT`)
* inout (`INOUT`)
* input-monitor (`INM`)
* output-monitor (`OUTM`)

>>> from tabulate import tabulate
>>> import ucdp as u
>>> directions = (u.IN, u.OUT, u.INOUT)
>>> overview = [(d, d.mode, d.name, d.suffix) for d in directions]
>>> print(tabulate(overview, ("Direction", ".mode", ".name", ".suffix")))
Direction      .mode  .name    .suffix
-----------  -------  -------  ---------
IN                 1  IN       _i
OUT               -1  OUT      _o
INOUT              0  INOUT    _io

Common Features
===============

You can calculate with :any:`Orientation` ...

>>> FWD * FWD
FWD
>>> FWD * BWD
BWD

... and :any:`Direction`

>>> IN * FWD
IN
>>> IN * BWD
OUT
>>> IN * FWD * BWD * FWD
OUT
>>> IN * BWD * BWD
IN
>>> OUT * FWD
OUT
>>> OUT * BWD
IN
>>> INOUT * FWD
INOUT
>>> INOUT * BWD
INOUT
>>> FWD * OUT
BWD

:any:`Orientation` and :any:`Direction` are singletons. There is just one instance of each.

>>> IN is IN
True
>>> IN is (IN * BWD * BWD)
True

:any:`Orientation` and :any:`Direction` can be compared:

>>> IN == IN
True
>>> IN == OUT
False
>>> IN == FWD
False

API
===
"""

from typing import ClassVar, Optional, Union

from .object import LightObject


class AOrientation(LightObject):
    """Abstract Orientation."""

    _NAMEMAP: ClassVar[dict[int, str]] = {}

    mode: int
    """
    Integer representation.
    """

    def __init__(self, mode):
        if mode not in self._NAMEMAP:
            raise ValueError(f"Invalid mode {mode}")
        super().__init__(mode=mode)

    @property
    def name(self):
        """Name."""
        return self._NAMEMAP[self.mode]

    def __str__(self):
        return self._NAMEMAP[self.mode]

    def __repr__(self):
        return self._NAMEMAP[self.mode]

    def __mul__(self, other) -> "AOrientation":
        if isinstance(other, AOrientation):
            return self.__class__(mode=self.mode * other.mode)
        return NotImplemented

    @property
    def suffix(self) -> str:
        """Suffix."""
        return ""

    @classmethod
    def cast(cls, value: Union["AOrientation", int]) -> "AOrientation":
        """Cast `value`."""
        if isinstance(value, cls):
            return value
        if isinstance(value, AOrientation):
            return cls(mode=value.mode)
        if isinstance(value, int) and value in cls._NAMEMAP.keys():
            return cls(mode=value)
        raise ValueError(f"Cannot cast {value}")


class Orientation(AOrientation):
    """
    Type Orientation.

    >>> import ucdp as u
    >>> u.Orientation.cast(1)
    FWD
    >>> u.Orientation.cast(u.FWD)
    FWD
    >>> u.Orientation.cast(u.IN)
    FWD
    """

    _NAMEMAP: ClassVar[dict[int, str]] = {
        1: "FWD",
        -1: "BWD",
    }


class Direction(AOrientation):
    """Port Direction."""

    _NAMEMAP: ClassVar[dict[int, str]] = {
        1: "IN",
        -1: "OUT",
        0: "INOUT",
    }

    _SUFFIXMAP: ClassVar[dict[int, str]] = {
        1: "_i",
        -1: "_o",
        0: "_io",
    }

    @property
    def suffix(self) -> str:
        """Suffix."""
        return self._SUFFIXMAP[self.mode]

    @staticmethod
    def from_name(name: str) -> Optional["Direction"]:
        """
        Determine :any:`Direction` by suffix of `name`.

        >>> Direction.from_name('ctrl_i')
        IN
        >>> Direction.from_name('ctrl_o')
        OUT
        >>> Direction.from_name('ctrl_io')
        INOUT
        >>> Direction.from_name('ctrl_s')
        >>> Direction.from_name('')
        """
        for mode, suffix in Direction._SUFFIXMAP.items():
            if name.endswith(suffix):
                return Direction(mode=mode)
        return None


FWD = Orientation(mode=1)
BWD = Orientation(mode=-1)

IN = Direction(mode=1)
OUT = Direction(mode=-1)
INOUT = Direction(mode=0)

DIRECTION_SUFFIXES = tuple(Direction._SUFFIXMAP.values())
"""Valid Direction Suffixes."""



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""UART Example."""

from fileliststandard import HdlFileList
from glbl_lib.bus import BusType  # (2)
from glbl_lib.clk_gate import ClkGateMod  # (3)
from glbl_lib.regf import RegfMod  # (4)

import ucdp as u  # (1)


class UartIoType(u.AStructType):
    """UART IO."""

    title: str = "UART"
    comment: str = "RX/TX"

    def _build(self) -> None:
        self._add("rx", u.BitType(), u.BWD)  # (5)
        self._add("tx", u.BitType(), u.FWD)  # (6)


class UartMod(u.AMod):
    """A Simple UART."""

    filelists: u.ClassVar[u.ModFileLists] = (
        HdlFileList(gen="full"),
        u.ModFileList(
            name="header",
            filepaths=("$PRJROOT/{mod.modname}.hpp"),
            template_filepaths=("hpp.mako"),
        ),
    )

    tags: u.ClassVar[u.ModTags] = {"intf"}

    def _build(self) -> None:
        self.add_port(u.ClkRstAnType(), "main_i")
        self.add_port(UartIoType(), "uart_i", route="create(u_core/uart_i)", clkrel=u.ASYNC)
        self.add_port(BusType(), "bus_i", clkrel="main_clk_i")

        clkgate = ClkGateMod(self, "u_clk_gate")
        clkgate.con("clk_i", "main_clk_i")
        clkgate.con("clk_o", "create(clk_s)")

        regf = RegfMod(self, "u_regf")
        regf.con("main_i", "main_i")
        regf.con("bus_i", "bus_i")

        core = UartCoreMod(parent=self, name="u_core")

        core.add_port(u.ClkRstAnType(), "main_i")
        core.con("main_clk_i", "clk_s")
        core.con("main_rst_an_i", "main_rst_an_i")
        core.con("create(regf_i)", "u_regf/regf_o")

        word = regf.add_word("ctrl")
        word.add_field("ena", u.EnaType(), is_readable=True, route="u_clk_gate/ena_i")
        word.add_field("strt", u.BitType(), is_writable=True, route="create(u_core/strt_i)")


class UartCoreMod(u.ACoreMod):
    """A Simple UART."""

    filelists: u.ClassVar[u.ModFileLists] = (HdlFileList(gen="inplace"),)



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Module Defines."""

from test2ref import assert_refdata

import ucdp as u


class IpMod(u.AMod):
    """IP."""

    def _build(self):
        self.add_port(u.BitType(), "d_i")
        self.add_port(u.BitType(), "e_i", ifdefs=("HAS_E", "!NO_EXTRA"))
        self.add_port(u.BitType(), "f_i", ifdefs=("!HAS_E", "!NO_EXTRA"))
        self.add_port(u.BitType(), "d_o")


class TopMod(u.AMod):
    """Top."""

    def _build(self):
        IpMod(self, "u_a", defines={"HAS_E": None})
        IpMod(self, "u_b")
        IpMod(self, "u_c", defines={"HAS_E": 1, "NO_EXTRA": 1})


def test_defines(tmp_path, capsys):
    """Test Defines."""
    top = TopMod()
    for inst in top.insts:
        print(inst)
        for port in inst.ports:
            print(" ", repr(port))

    assert_refdata(test_defines, tmp_path, capsys=capsys)



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Multiplexer."""

from collections import defaultdict
from collections.abc import Iterator
from functools import cached_property

from .assigns import Assign, Assigns, Drivers
from .doc import Doc
from .expr import Expr
from .exprparser import ExprParser, Parseable
from .ident import Ident, Idents
from .object import Field, NamedObject, computed_field
from .signal import BaseSignal
from .typebase import BaseScalarType

MuxBranch = dict[Expr, Assigns]


class Mux(NamedObject):
    """
    Multiplexer.

    Attributes:
        name: Name
        targets: Ports and signals allowed to be assigned
        namespace: Ports, signals, parameter and local parameter as source
        drivers: Drivers, for multiple-driver tracking.
        parser: Parser
        doc: Documentation
    """

    targets: Idents = Field(repr=False)
    namespace: Idents = Field(repr=False)
    drivers: Drivers = Field(repr=False, default_factory=dict)
    parser: ExprParser = Field(repr=False)
    doc: Doc = Doc()

    @computed_field(repr=False)
    @cached_property
    def __assigns(self) -> Assigns:
        return Assigns(targets=self.targets, sources=self.namespace, drivers=self.drivers)

    @computed_field(repr=False)
    @cached_property
    def __mux(self) -> dict[Expr, MuxBranch]:
        return defaultdict(dict)

    def set(self, sel: Parseable, cond: Parseable, out: Parseable, value: Parseable):
        """
        Set Multiplexer.

        Args:
            sel: Select Signal
            cond: Condition to be met on `sel`
            out: Output to be assigned
            value: Value.
        """
        parse = self.parser.parse
        selexpr = parse(sel, types=BaseScalarType)
        condexpr = parse(cond)
        outexpr: BaseSignal = parse(out, only=BaseSignal)  # type: ignore[assignment]
        valueexpr = parse(value)
        mux = self.__mux
        # Track Assignments
        self.__assigns.set(outexpr, valueexpr, overwrite=True)
        # Mux Assignments
        try:
            condassigns = mux[selexpr][condexpr]
        except KeyError:
            mux[selexpr][condexpr] = condassigns = Assigns(targets=self.targets, sources=self.namespace)
        condassigns.set(outexpr, valueexpr)

    def set_default(self, out: Parseable, value: Parseable):
        """
        Set Multiplexer.

        Args:
            out: Output to be assigned
            value: Value.
        """
        parse = self.parser.parse
        outexpr: BaseSignal = parse(out, only=BaseSignal)  # type: ignore[assignment]
        valueexpr = parse(value)
        self.__assigns.set_default(outexpr, valueexpr)

    def defaults(self) -> Iterator[Assign]:
        """Defaults."""
        return self.__assigns.defaults()

    def __iter__(self) -> Iterator[tuple[Expr, MuxBranch]]:  # type: ignore[override]
        yield from self.__mux.items()

    @property
    def sels(self) -> tuple[Ident, ...]:
        """Selects."""
        return tuple(self.__mux.keys())  # type: ignore[arg-type]
